"""Renderer for performing render operations for the pod."""

import sys
from multiprocessing.dummy import Pool as ThreadPool
from grow.pods import errors


class Error(Exception):
    """Base renderer error."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class RenderError(Error):
    """Errors that occured during the rendering."""

    def __init__(self, message, err, err_tb):
        super(RenderError, self).__init__(message)
        self.err = err
        self.err_tb = err_tb


class RenderNotStartedError(Error):
    """Rendering was not started."""
    pass


def load_func(batch, source_dir, tick=None):
    """Render the controller."""
    result = LoadBatchResult()
    for item in batch:
        controller = item['controller']
        try:
            result.loaded_docs.append(controller.load(source_dir))
        except errors.BuildError as err:
            result.load_errors.append(RenderError(
                "Error loading {}".format(controller.serving_path),
                err, err.traceback))
        except Exception as err:  # pylint: disable=broad-except
            _, _, err_tb = sys.exc_info()
            result.load_errors.append(RenderError(
                "Error loading {}".format(controller.serving_path),
                err, err_tb))
        finally:
            if tick:
                tick()
    return result


def render_func(batch, tick=None):
    """Render the controller."""
    result = RenderBatchResult()
    for item in batch:
        controller = item['controller']
        try:
            result.rendered_docs.append(controller.render(
                jinja_env=item['jinja_env']))
        except errors.BuildError as err:
            result.render_errors.append(RenderError(
                "Error rendering {}".format(controller.serving_path),
                err, err.traceback))
        except Exception as err:  # pylint: disable=broad-except
            _, _, err_tb = sys.exc_info()
            result.render_errors.append(RenderError(
                "Error rendering {}".format(controller.serving_path),
                err, err_tb))
        finally:
            if tick:
                tick()
    return result


class RenderBatches(object):
    """Handles the batching of rendering."""

    def __init__(self, render_pool, profile, tick=None, batch_size=None):
        self.batch_size = batch_size
        self.profile = profile
        self.render_pool = render_pool
        self.tick = tick
        self._batches = {}

    def __len__(self):
        count = 0
        for _, batch in self._batches.items():
            count = count + len(batch)
        return count

    def _get_batch(self, locale):
        if locale not in self._batches:
            self._batches[locale] = RenderLocaleBatch(
                self.render_pool.get_jinja_env(locale), self.profile, tick=self.tick,
                batch_size=self.batch_size)
        return self._batches[locale]

    def add(self, controller, *args, **kwargs):
        """Add an item to be rendered to the batch."""
        batch = self._get_batch(controller.locale)
        batch.add(controller, *args, **kwargs)

    def load(self, source_dir, use_threading=True):
        """Load all of the batches."""
        load_errors = []
        load_docs = []

        # Disable threaded rendering until it can be fixed.
        use_threading = False

        if not ThreadPool or not use_threading:
            for _, batch in self._batches.items():
                batch_docs, batch_errors = batch.load_sync(source_dir)
                load_errors = load_errors + batch_errors
                load_docs = load_docs + batch_docs
        else:
            for _, batch in self._batches.items():
                batch.load_start(source_dir)
            for _, batch in self._batches.items():
                batch_docs, batch_errors = batch.load_finish()
                load_errors = load_errors + batch_errors
                load_docs = load_docs + batch_docs

        return load_docs, load_errors

    def render(self, use_threading=True):
        """Render all of the batches."""
        render_errors = []
        rendered_docs = []

        # Disable threaded rendering until it can be fixed.
        use_threading = False

        if not ThreadPool or not use_threading:
            for _, batch in self._batches.items():
                batch_docs, batch_errors = batch.render_sync()
                render_errors = render_errors + batch_errors
                rendered_docs = rendered_docs + batch_docs
        else:
            for _, batch in self._batches.items():
                batch.render_start()
            for _, batch in self._batches.items():
                batch_docs, batch_errors = batch.render_finish()
                render_errors = render_errors + batch_errors
                rendered_docs = rendered_docs + batch_docs

        return rendered_docs, render_errors


class RenderLocaleBatch(object):
    """Handles the rendering and threading of the controllers."""

    BATCH_DEFAULT_SIZE = 300  # Default number of documents in a batch.

    def __init__(self, jinja_env, profile, tick=None, batch_size=None):
        self.batch_size = batch_size or self.BATCH_DEFAULT_SIZE
        self.jinja_env = jinja_env
        self.profile = profile
        self.tick = tick
        self.batches = [[]]
        self._is_loading = False
        self._is_rendering = False
        self._results = None
        self._thread_pool = None

    def __len__(self):
        count = 0
        for batch in self.batches:
            count = count + len(batch)
        return count

    def _get_batch(self):
        # Ensure that batch is not over the max size.
        batch = self.batches[len(self.batches) - 1]
        if len(batch) >= self.batch_size:
            self.batches.append([])
            batch = self.batches[len(self.batches) - 1]
        return batch

    def add(self, controller, *args, **kwargs):
        """Add an item to be rendered to the batch."""
        batch = self._get_batch()

        batch.append({
            'controller': controller,
            'jinja_env': self.jinja_env,
            'args': args,
            'kwargs': kwargs,
        })

    def load_start(self, source_dir):
        """Start the batches loading."""
        self._thread_pool = ThreadPool(len(self.batches))
        self._results = self._thread_pool.imap_unordered(
            load_func, self.batches, source_dir)
        self._is_loading = True

    def load_finish(self):
        """Finish in progress batches loading."""
        if not self._is_loading:
            raise RenderNotStartedError('Rendering was never started')

        load_errors = []
        loaded_docs = []

        for batch_result in self._results:
            load_errors = load_errors + batch_result.load_errors
            loaded_docs = loaded_docs + batch_result.loaded_docs
            if self.tick:
                for _ in batch_result.load_errors:
                    self.tick()
                for _ in batch_result.loaded_docs:
                    self.tick()
            for result in batch_result.loaded_docs:
                self.profile.add_timer(result.load_timer)

        self._thread_pool.close()
        self._thread_pool.join()
        self._is_loading = False

        return loaded_docs, load_errors

    def load_sync(self, source_dir):
        """Syncronous loading for non-threaded loading."""
        load_errors = []
        loaded_docs = []

        for batch in self.batches:
            batch_result = load_func(batch, source_dir, tick=self.tick)
            load_errors = load_errors + batch_result.load_errors
            loaded_docs = loaded_docs + batch_result.loaded_docs

        return loaded_docs, load_errors

    def render_start(self):
        """Start the batches rendering."""
        self._thread_pool = ThreadPool(len(self.batches))
        self._results = self._thread_pool.imap_unordered(
            render_func, self.batches)
        self._is_rendering = True

    def render_finish(self):
        """Finish in progress batches rendering."""
        if not self._is_rendering:
            raise RenderNotStartedError('Rendering was never started')

        render_errors = []
        rendered_docs = []

        for batch_result in self._results:
            render_errors = render_errors + batch_result.render_errors
            rendered_docs = rendered_docs + batch_result.rendered_docs
            if self.tick:
                for _ in batch_result.render_errors:
                    self.tick()
                for _ in batch_result.rendered_docs:
                    self.tick()
            for result in batch_result.rendered_docs:
                self.profile.add_timer(result.render_timer)

        self._thread_pool.close()
        self._thread_pool.join()
        self._is_rendering = False

        return rendered_docs, render_errors

    def render_sync(self):
        """Syncronous rendering for non-threaded rendering."""
        render_errors = []
        rendered_docs = []

        for batch in self.batches:
            batch_result = render_func(batch, tick=self.tick)
            render_errors = render_errors + batch_result.render_errors
            rendered_docs = rendered_docs + batch_result.rendered_docs

        return rendered_docs, render_errors


class LoadBatchResult(object):
    """Results from a batched loading."""

    def __init__(self):
        self.load_errors = []
        self.loaded_docs = []


class RenderBatchResult(object):
    """Results from a batched rendering."""

    def __init__(self):
        self.render_errors = []
        self.rendered_docs = []



from . import base
from googleapiclient import discovery
from googleapiclient import errors
from grow.common import oauth
from grow.common import utils
from protorpc import messages
import datetime
import httplib2
import logging


OAUTH_SCOPE = 'https://www.googleapis.com/auth/blogger'


# Silence extra logging from googleapiclient.
discovery.logger.setLevel(logging.WARNING)


class BloggerPreprocessor(base.BasePreprocessor):
    _edit_url_format = 'https://www.blogger.com/blogger.g?blogID={blog_id}#allposts'
    KIND = 'blogger'

    class Config(messages.Message):
        blog_id = messages.IntegerField(1)
        collection = messages.StringField(2)
        preserve = messages.StringField(3, default='builtins')
        markdown = messages.BooleanField(4, default=False)
        authenticated = messages.BooleanField(5, default=True)

    @staticmethod
    def create_service(authenticated=True):
        http = httplib2.Http()
        if authenticated:
            credentials = oauth.get_or_create_credentials(
                scope=OAUTH_SCOPE, storage_key='Grow SDK')
            http = credentials.authorize(http)
        key = None if authenticated else oauth.BROWSER_API_KEY
        return discovery.build('blogger', 'v3', http=http, developerKey=key)

    def get_edit_url(self, doc=None):
        """Returns the URL to edit in Blogger."""
        return BloggerPreprocessor._edit_url_format.format(
            blog_id=self.config.blog_id)

    def run(self, build=True):
        self.execute()

    def execute(self):
        items = BloggerPreprocessor.download_items(
            blog_id=self.config.blog_id,
            logger=self.pod.logger,
            authenticated=self.config.authenticated)
        self.bind_collection(items, self.config.collection)

    def _parse_item(self, item):
        item_id = item.pop('id')
        ext = 'md' if self.config.markdown else 'html'
        basename = '{}.{}'.format(item_id, ext)
        body = item.pop('content')
        fields = item
        # Formatted like: 2011-05-20T11:45:23-07:00
        published = fields['published'][:-6]
        published_dt = datetime.datetime.strptime(published, '%Y-%m-%dT%H:%M:%S')
        fields['$date'] = published_dt
        if 'title' in fields:
            fields['$title'] = fields.pop('title')
        if self.config.markdown:
            body = utils.clean_html(body, convert_to_markdown=True)
        return fields, body, basename

    @classmethod
    def download_items(cls, blog_id, logger=None, authenticated=True):
        logger = logger or logging
        service = cls.create_service(authenticated=authenticated)
        items = []
        # pylint: disable=no-member
        posts = service.posts()
        request = posts.list(blogId=blog_id)
        while request != None:
            posts_doc = request.execute()
            if 'items' in posts_doc and not (posts_doc['items'] is None):
                items.extend(posts_doc['items'])
            elif not items:
                logger.error('Received: {}'.format(posts_doc))
                text = 'Unable to download Blogger blog: {}'
                raise base.PreprocessorError(text.format(blog_id))
            request = posts.list_next(request, posts_doc)
        return items

    @classmethod
    def download_item(cls, blog_id, post_id, authenticated=True):
        service = cls.create_service(authenticated=authenticated)
        # pylint: disable=no-member
        resp = service.posts().get(blogId=blog_id, postId=post_id).execute()
        return resp

    def bind_collection(self, items, collection_pod_path):
        collection = self.pod.get_collection(collection_pod_path)
        existing_pod_paths = [
            doc.pod_path for doc in collection.list_docs(recursive=False, inject=False)]
        new_pod_paths = []
        for item in items:
            fields, body, basename = self._parse_item(item)
            doc = collection.create_doc(basename, fields=fields, body=body)
            new_pod_paths.append(doc.pod_path)
            self.pod.logger.info('Saved -> {}'.format(doc.pod_path))
        pod_paths_to_delete = set(existing_pod_paths) - set(new_pod_paths)
        for pod_path in pod_paths_to_delete:
            self.pod.delete_file(pod_path)
            self.pod.logger.info('Deleted -> {}'.format(pod_path))

    def _normalize_path(self, path):
        """Normalizes a collection path."""
        return path.rstrip('/')

    def can_inject(self, doc=None, collection=None):
        if not self.injected:
            return False
        if doc and doc.pod_path.startswith(self.config.collection):
            return True
        if (collection and
                self._normalize_path(collection.pod_path)
                == self._normalize_path(self.config.collection)):
            return True
        return False

    def inject(self, doc):
        path = doc.pod_path
        post_id = doc.base
        try:
            item = BloggerPreprocessor.download_item(
                blog_id=self.config.blog_id,
                post_id=post_id,
                authenticated=self.config.authenticated)
        except (errors.HttpError, base.PreprocessorError):
            text = 'Error downloading Blogger post -> {}'.format(path)
            raise base.PreprocessorError(text)
        if not item:
            return
        fields, body, _ = self._parse_item(item)
        if doc.exists:
            existing_data = doc.format.front_matter.data
        else:
            existing_data = {}
        fields = utils.format_existing_data(
            old_data=existing_data, new_data=fields,
            preserve=self.config.preserve)
        doc.inject(fields=fields, body=body)
        return self



"""Grow extensions installer class."""

import subprocess
from grow.sdk.installers import base_installer


INSTALL_HASH_FILE_NAME = 'install-hash.txt'


class ExtensionsInstaller(base_installer.BaseInstaller):
    """Grow extensions installer."""

    KIND = 'extensions'

    def _generate_hashes(self):
        """Create hash string based on the package and lock files."""
        if not self.pod.file_exists(self.extensions_filename):
            return ''
        return self.pod.hash_file(self.extensions_filename)

    @property
    def post_install_messages(self):
        """List of messages to display after installing."""
        return ['Finished: Extensions -> {}/'.format(self.pod.extensions_dir)]

    @property
    def extensions_filename(self):
        """Filename for the install hash."""
        return '/{}'.format(self.pod.FILE_EXTENSIONS)

    @property
    def install_hash_filename(self):
        """Filename for the install hash."""
        return '/{}/{}'.format(self.pod.extensions_dir, INSTALL_HASH_FILE_NAME)

    @property
    def should_run(self):
        """Should the installer run?"""
        if not self.pod.file_exists(self.extensions_filename):
            return False

        if self.config.get('force', None):
            return True

        if self.pod.file_exists(self.install_hash_filename):
            existing_hash = self.pod.read_file(self.install_hash_filename)
            current_hash = self._generate_hashes()
            if existing_hash == current_hash:
                self.pod.logger.info('Skipping extensions install since no changes detected.')
                self.pod.logger.info('   Use `grow install -f` to force install.')
                return False
        return True

    def check_prerequisites(self):
        """Check if required prerequisites are installed or available."""
        status_command = 'pip3 --version > /dev/null 2>&1'
        not_found = subprocess.call(
            status_command, **self.subprocess_args(shell=True)) == 127
        if not_found:
            raise base_installer.MissingPrerequisiteError(
                'The `pip3` command was not found.')

    def install(self):
        """Install dependencies."""
        extensions_dir = self.pod.extensions_dir
        init_file_name = '/{}/__init__.py'.format(extensions_dir)
        if not self.pod.file_exists(init_file_name):
            self.pod.write_file(init_file_name, '')
        install_command = 'pip3 install -U -q -t {} -r {}'.format(
            extensions_dir, self.pod.FILE_EXTENSIONS)
        process = subprocess.Popen(
            install_command, **self.subprocess_args(shell=True))
        code = process.wait()
        if not code:
            self.pod.write_file(self.install_hash_filename, self._generate_hashes())
            return
        raise base_installer.InstallError(
            'There was an error running `{}`.'.format(install_command))



"""Utility for dealing with sub processes."""

import time

def kill_child_processes(child_processes, timeout=2):
    """Kill all subprocesses if they timeout."""
    start_time = time.time()

    has_pending = True
    while has_pending:
        has_pending = False
        for process in child_processes:
            try:
                if process.poll() is None:
                    has_pending = True
                    process.terminate()
            except OSError:
                # Ignore the error.  The OSError doesn't seem to be documented(?)
                pass

        if not has_pending:
            return

        elapsed_time = time.time() - start_time
        if elapsed_time > timeout:
            for process in child_processes:
                if process.poll() is None:
                    process.kill()
            return



"""Base class for destinations.

A destination is a place where Grow static builds can be deployed. Grow takes
a pod, builds a static fileset, and deploys it to a remote location (the
desitnation), suitable for serving the web site to live end users.

A "deployment" is a destination loaded up with a configuration.

Currently, Grow only supports static deployments, however, this may change as
Grow implements features such as "password protection" and geolocation.

The deployment process generally works like this:

  (1) A pod is exported, creating a dictionary mapping file paths to content.
  (2) A connection is made between Grow and the destination.
  (3) Control files are retrieved from the destination, if they exist. All
      control files are serialized ProtoRPC messages. The most important
      control file is "index.proto.json", which contains an index of file
      paths to sha-1 hashes of each file's content.
  (4) An index is generated locally, and the local index is compared to the
      index at the destination. This allows Grow to produce a diff between
      the local ("canary") fileset and the destination's fileset.
  (5) An integration test (if any) is performed.
  (6) If the deployment is a dry run, the process ends here.
  (7) Any required pre-launch configuration to the destination is applied.
  (8) The diff between the local and remote fileset is applied.
  (9) Updated control files are written to the desination.
  (10) Any required post-launch configuration to the destination is applied.

  The deployment is complete!

All deployments follow this process, and the BaseDestination class takes
care of most of the hard work and business logic. So if you're adding a new
destination, you'll just have to implement the following methods/properties:

  delete_file(self, path)
    Deletes a file at the destination, given the file's pod path.

  read_file(self, path)
    Reads a file at the destination, returning the file's content.

  write_file(self, rendered_doc)
    Writes a file at the destination, given the file's rendered_document.

  write_files(self, paths_to_rendered_doc)
    Writes files in bulk, given a mapping of paths to rendered_document.

  KIND
    A string identifying the deployment.

The following methods are optional to implement:

  postlaunch(self, dry_run)
    Performs any post-launch tasks.

  prelaunch(self, dry_run)
    Performs any pre-launch configuration/tasks.

  write_control_file(self, basename, content):
    Writes a control file to the destination.

If your destination requires configuration, you should add a nested class:

  Config
    A ProtoRPC message for destination configuration.

New builtin destinations should be added to the list of builtins in
grow/deployments/deployments.py. Proprietary destinations can be registered
using deployments.register_destination.
"""

import inspect
import io
import json
import logging
import os
import subprocess
import re
import sys
from grow.common import utils
from grow.deployments import indexes
from grow.deployments import tests
from grow.performance import profile_report
from grow.pods import env
from grow.pods import pods
from grow.rendering import rendered_document
from . import messages


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class CommandError(Error):
    pass


class DestinationTestCase(object):

    def __init__(self, deployment):
        self.deployment = deployment

    def test_write_file(self):
        basename = 'test.tmp'
        title = 'Can write files to {}'.format(self.deployment)
        message = messages.TestResultMessage(title=title)
        self.deployment.write_control_file(basename, 'test')
        content = self.deployment.read_control_file(basename)
        if content != 'test':
            message.result = messages.Result.FAIL
        self.deployment.delete_control_file(basename)
        return message

    def __iter__(self):
        for name, func in inspect.getmembers(self, predicate=inspect.ismethod):
            if name.startswith('test_'):
                yield func


class BaseDestination(object):
    """Base destination for building and deploying."""

    TestCase = DestinationTestCase
    diff_basename = 'diff.proto.json'
    index_basename = 'index.proto.json'
    stats_basename = 'stats.proto.json'
    threaded = True
    batch_writes = False
    success = False

    def __init__(self, config, name='default'):
        self.config = config
        self.name = name
        self.pod = None
        self._diff = None
        self._confirm = None

    def __str__(self):
        return self.__class__.__name__

    @property
    def control_dir(self):
        if self.config.keep_control_dir:
            control_dir = pods.Pod.PATH_CONTROL
            return os.path.join(self.pod.root, control_dir, 'deployments', self.name)
        if self._has_custom_control_dir:
            return self.config.control_dir
        return pods.Pod.PATH_CONTROL

    @property
    def filters(self):
        """Configuration for deployment filters."""
        try:
            return self.config.filters
        except AttributeError:
            return []

    @property
    def prevent_untranslated(self):
        """Configuration for untranslated messages preventing deployment."""
        if not self.config.base_config:
            return False
        return self.config.base_config.prevent_untranslated

    @property
    def storage(self):
        raise NotImplementedError

    def _get_remote_index(self):
        try:
            content = self.read_control_file(self.index_basename)
            return indexes.Index.from_string(content)
        except IOError:
            logging.info('Unable to find remote index: {}'.format(
                self.index_basename))
            return indexes.Index.create()

    def export_profile_report(self):
        """Write the results of the profiling timers."""
        report = profile_report.ProfileReport(self.pod.profile)
        file_name = '{}profile.json'.format(self.control_dir)
        self.pod.write_file(file_name, json.dumps(report.export()))
        logging.info('Profiling data exported to {}'.format(file_name))

    def export_untranslated_catalogs(self):
        dir_path = '{}untranslated/'.format(self.control_dir)
        catalogs = self.pod.translation_stats.export_untranslated_catalogs(
            self.pod, dir_path=dir_path)
        self.pod.delete_files([dir_path], recursive=True,
                              pattern=re.compile(r'\.po$'))
        for _, catalog in catalogs.items():
            catalog.save()
        if self.pod.translation_stats.stacktraces:
            self.pod.write_file(os.path.join(dir_path, 'tracebacks.log'),
                self.pod.translation_stats.export_untranslated_tracebacks())
        if catalogs or self.pod.translation_stats.stacktraces:
            logging.info('Untranslated strings exported to {}'.format(dir_path))

    def get_env(self):
        """Returns an environment object based on the config."""
        if self.config.env:
            return env.Env(self.config.env)
        config = env.EnvConfig(host='localhost')
        return env.Env(config)

    def read_file(self, path):
        """Returns a file-like object."""
        raise NotImplementedError

    def write_files(self, paths_to_rendered_doc):
        """Writes files in bulk."""
        raise NotImplementedError

    def write_file(self, rendered_doc):
        """Writes an individual file."""
        raise NotImplementedError

    def delete_file(self, path):
        raise NotImplementedError

    @property
    def _has_custom_control_dir(self):
        return (hasattr(self.config, 'control_dir')
                and self.config.control_dir is not None)

    def delete_control_file(self, path):
        path = os.path.join(self.control_dir, path.lstrip('/'))
        if self.config.keep_control_dir:
            return self.pod.delete_file(path)
        if self._has_custom_control_dir:
            return self.storage.delete(path)
        return self.delete_file(path)

    def read_control_file(self, path):
        path = os.path.join(self.control_dir, path.lstrip('/'))
        if self.config.keep_control_dir:
            return self.pod.read_file(path)
        if self._has_custom_control_dir:
            return self.storage.read(path)
        return self.read_file(path)

    def write_control_file(self, path, content):
        path = os.path.join(self.control_dir, path.lstrip('/'))
        if self.config.keep_control_dir:
            return self.pod.write_file(path, content)
        if self._has_custom_control_dir:
            return self.storage.write(path, content)
        if self.batch_writes:
            return self.write_files({
                path: rendered_document.RenderedDocument(path, content)
            })
        return self.write_file(
            rendered_document.RenderedDocument(path, content))

    def test(self):
        results = messages.TestResultsMessage(test_results=[])
        failures = []
        test_case = self.TestCase(self)
        for func in test_case:
            result_message = func()
            results.test_results.append(result_message)
            if result_message.result == messages.Result.FAIL:
                failures.append(result_message)
        tests.print_results(results)
        return results

    def postlaunch(self, dry_run=False):
        pass

    def prelaunch(self, dry_run=False):
        pass

    def login(self, account, reauth=False):
        pass

    def dump(self, pod, pod_paths=None, use_threading=True, source_dir=None):
        """Dump the contents of the pod."""
        pod.set_env(self.get_env())
        return pod.dump(
            pod_paths=pod_paths, use_threading=use_threading,
            source_dir=source_dir)

    def deploy(self, content_generator, stats=None, repo=None, dry_run=False,
               confirm=False, test=True, is_partial=False, require_translations=False):
        self._confirm = confirm
        self.prelaunch(dry_run=dry_run)
        if test:
            self.test()
        with self.pod.profile.timer('deployments.destination.Base.deploy'):
            try:
                deployed_index = self._get_remote_index()
                if require_translations:
                    self.pod.enable(self.pod.FEATURE_TRANSLATION_STATS)
                diff, new_index, paths_to_rendered_doc = indexes.Diff.stream(
                    deployed_index, content_generator, repo=repo, is_partial=is_partial)
                self._diff = diff
                if indexes.Diff.is_empty(diff):
                    logging.info('Finished with no diffs since the last build.')
                    return
                if dry_run:
                    return
                indexes.Diff.pretty_print(diff)
                if require_translations and self.pod.translation_stats.untranslated:
                    self.pod.translation_stats.pretty_print()
                    raise pods.Error('Aborted deploy due to untranslated strings. '
                                     'Use the --force-untranslated flag to force deployment.')
                if confirm:
                    text = 'Proceed to deploy? -> {}'.format(self)
                    if not utils.interactive_confirm(text):
                        logging.info('Aborted.')
                        return
                indexes.Diff.apply(
                    diff, paths_to_rendered_doc, write_func=self.write_file,
                    batch_write_func=self.write_files, delete_func=self.delete_file,
                    threaded=self.threaded, batch_writes=self.batch_writes)
                self.write_control_file(
                    self.index_basename, indexes.Index.to_string(new_index))
                if stats is not None:
                    self.write_control_file(self.stats_basename, stats.to_string())
                else:
                    self.delete_control_file(self.stats_basename)
                if diff:
                    self.write_control_file(
                        self.diff_basename, indexes.Diff.to_string(diff))
                self.success = True
            finally:
                self.postlaunch()
        return diff

    def command(self, command):
        with io.BytesIO() as fp:
            proc = subprocess.Popen(command, stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, shell=True)
            for line in iter(proc.stdout.readline, ''):
                sys.stdout.write(line)
                fp.write(line)
            err = proc.stderr.read()
            if err:
                raise CommandError(err)



"""Controller for rendering pod content."""

import datetime
import mimetypes
import os
import sys
import time
from grow.common import utils
from grow.documents import static_document
from grow.pods import errors
from grow.rendering import rendered_document
from grow.templates import doc_dependency
from grow.templates import tags


class Error(Exception):
    """Base rendering pool error."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class UnknownKindError(Error):
    """Unknown kind of information."""
    pass


class IgnoredPathError(Error):
    """Document is being served at an ignored path."""
    pass


class RenderController(object):
    """Controls how the content is rendered and evaluated."""

    def __init__(self, pod, serving_path, route_info, params=None, is_threaded=False):
        self.pod = pod
        self.serving_path = serving_path
        self.route_info = route_info
        self.params = params if params is not None else {}
        self.render_timer = None
        self.use_jinja = False
        self.is_threaded = is_threaded

    @staticmethod
    def clean_source_dir(source_dir):
        """Clean the formatting of the source dir to format correctly."""
        source_dir = source_dir.strip()
        source_dir = source_dir.rstrip(os.path.sep)
        return source_dir

    @staticmethod
    def from_route_info(pod, serving_path, route_info, params=None):
        """Create the correct controller based on the route info."""
        if params is None:
            params = {}
        if route_info.kind == 'doc':
            return RenderDocumentController(
                pod, serving_path, route_info, params=params)
        elif route_info.kind == 'static':
            return RenderStaticDocumentController(
                pod, serving_path, route_info, params=params)
        elif route_info.kind == 'sitemap':
            return RenderSitemapController(
                pod, serving_path, route_info, params=params)
        elif route_info.kind == 'error':
            return RenderErrorController(
                pod, serving_path, route_info, params=params)
        raise UnknownKindError(
            'Do not have a controller for: {}'.format(route_info.kind))

    @property
    def locale(self):
        """Locale to use for rendering."""
        return None

    @property
    def mimetype(self):
        """Guess the mimetype of the content."""
        return 'text/plain'

    def get_http_headers(self):
        """Determine headers to serve for https requests."""
        headers = {}
        mimetype = self.mimetype
        if mimetype:
            headers['Content-Type'] = mimetype
        return headers

    def load(self, source_dir):
        """Load the pod content from file system."""
        raise NotImplementedError

    def render(self, jinja_env, request=None):
        """Render the pod content."""
        raise NotImplementedError

    def validate_path(self, *path_filters):
        """Validate that the path is valid against all filters."""
        # Default test against the pod filter for deployment specific filtering.
        path_filters = list(path_filters) or [self.pod.path_filter]
        for path_filter in path_filters:
            if not path_filter.is_valid(self.serving_path):
                text = '{} is an ignored path.'
                raise errors.RouteNotFoundError(text.format(self.serving_path))


class RenderDocumentController(RenderController):
    """Controller for handling rendering for documents."""

    def __init__(self, pod, serving_path, route_info, params=None, is_threaded=False):
        super(RenderDocumentController, self).__init__(
            pod, serving_path, route_info, params=params, is_threaded=is_threaded)
        self._doc = None
        self.use_jinja = True

    def __repr__(self):
        return '<RenderDocumentController({})>'.format(self.route_info.meta['pod_path'])

    @property
    def doc(self):
        """Doc for the controller."""
        if not self._doc:
            pod_path = self.route_info.meta['pod_path']
            locale = self.route_info.meta.get(
                'locale', self.params.get('locale'))
            self._doc = self.pod.get_doc(pod_path, locale=locale)
        return self._doc

    @property
    def locale(self):
        """Locale to use for rendering."""
        if 'locale' in self.route_info.meta:
            return self.route_info.meta['locale']
        return self.doc.locale if self.doc else None

    @property
    def mimetype(self):
        """Determine headers to serve for https requests."""
        return mimetypes.guess_type(self.doc.view)[0]

    @property
    def pod_path(self):
        """Locale to use for rendering."""
        if 'pod_path' in self.route_info.meta:
            return self.route_info.meta['pod_path']
        return self.doc.pod_path if self.doc else None

    @property
    def suffix(self):
        """Determine headers to serve for https requests."""
        _, ext = os.path.splitext(self.doc.view)
        if ext == '.html':
            return 'index.html'
        return ''

    def load(self, source_dir):
        """Load the pod content from file system."""
        timer = self.pod.profile.timer(
            'RenderDocumentController.load',
            label='{} ({})'.format(self.pod_path, self.locale),
            meta={
                'path': self.pod_path,
                'locale': str(self.locale)}
        ).start_timer()

        source_dir = self.clean_source_dir(source_dir)

        # Validate the path with the config filters.
        self.validate_path()

        try:
            doc = self.doc
            serving_path = self.serving_path
            if serving_path.endswith('/'):
                serving_path = '{}{}'.format(serving_path, self.suffix)

            rendered_path = '{}{}'.format(source_dir, serving_path)
            rendered_content = self.pod.storage.read(rendered_path)

            rendered_doc = rendered_document.RenderedDocument(
                serving_path, rendered_content)
            timer.stop_timer()
            return rendered_doc
        except Exception as err:
            exception = errors.BuildError(str(err))
            exception.traceback = sys.exc_info()[2]
            exception.controller = self
            exception.exception = err
            raise exception

    def render(self, jinja_env=None, request=None):
        """Render the document using the render pool."""
        timer = self.pod.profile.timer(
            'RenderDocumentController.render',
            label='{} ({})'.format(self.doc.pod_path, self.doc.locale),
            meta={
                'path': self.doc.pod_path,
                'locale': str(self.doc.locale)}
        ).start_timer()

        # Validate the path with the config filters.
        self.validate_path()

        doc = self.doc
        template = jinja_env['env'].get_template(doc.view.lstrip('/'))
        track_dependency = doc_dependency.DocDependency(doc)
        local_tags = tags.create_builtin_tags(
            self.pod, doc, track_dependency=track_dependency)
        # NOTE: This should be done using get_template(... globals=...)
        # or passed as an argument into render but
        # it is not available included inside macros???
        # See: https://github.com/pallets/jinja/issues/688
        template.globals['g'] = local_tags

        # Track the message stats, including untranslated strings.
        if self.pod.is_enabled(self.pod.FEATURE_TRANSLATION_STATS):
            template.globals['_'] = tags.make_doc_gettext(doc)

        try:
            doc.footnotes.reset()
            serving_path = doc.get_serving_path()
            if serving_path.endswith('/'):
                serving_path = '{}{}'.format(serving_path, self.suffix)

            content = self.pod.extensions_controller.trigger('pre_render', doc, doc.body)
            if content:
                doc.format.update(content=content)

            rendered_content = template.render({
                'doc': doc,
                'request': request,
                'env': self.pod.env,
                'podspec': self.pod.podspec,
                '_track_dependency': track_dependency,
            }).lstrip()
            rendered_content = self.pod.extensions_controller.trigger(
                'post_render', doc, rendered_content)
            rendered_doc = rendered_document.RenderedDocument(
                serving_path, rendered_content)
            timer.stop_timer()
            return rendered_doc
        except Exception as err:
            exception = errors.BuildError(str(err))
            exception.traceback = sys.exc_info()[2]
            exception.controller = self
            exception.exception = err
            raise exception


class RenderErrorController(RenderController):
    """Controller for handling rendering for errors."""

    def __init__(self, pod, serving_path, route_info, params=None, is_threaded=False):
        super(RenderErrorController, self).__init__(
            pod, serving_path, route_info, params=params, is_threaded=is_threaded)
        self.use_jinja = True

    def __repr__(self):
        return '<RenderErrorController({})>'.format(self.route_info.meta['view'])

    def load(self, source_dir):
        """Load the pod content from file system."""
        timer = self.pod.profile.timer(
            'RenderErrorController.load',
            label='{} ({})'.format(
                self.route_info.meta['key'], self.route_info.meta['view']),
            meta={
                'key': self.route_info.meta['key'],
                'view': self.route_info.meta['view'],
            }
        ).start_timer()

        source_dir = self.clean_source_dir(source_dir)

        # Validate the path with the config filters.
        self.validate_path()

        try:
            serving_path = '/{}.html'.format(self.route_info.meta['key'])
            rendered_path = '{}{}'.format(source_dir, serving_path)
            rendered_content = self.pod.storage.read(rendered_path)
            rendered_doc = rendered_document.RenderedDocument(
                serving_path, rendered_content)
            timer.stop_timer()
            return rendered_doc
        except Exception as err:
            text = 'Error building {}: {}'
            if self.pod:
                self.pod.logger.exception(text.format(self, err))
            exception = errors.BuildError(text.format(self, err))
            exception.traceback = sys.exc_info()[2]
            exception.controller = self
            exception.exception = err
            raise exception

    def render(self, jinja_env=None, request=None):
        """Render the document using the render pool."""
        timer = self.pod.profile.timer(
            'RenderErrorController.render',
            label='{} ({})'.format(
                self.route_info.meta['key'], self.route_info.meta['view']),
            meta={
                'key': self.route_info.meta['key'],
                'view': self.route_info.meta['view'],
            }
        ).start_timer()

        # Validate the path with the config filters.
        self.validate_path()

        with jinja_env['lock']:
            template = jinja_env['env'].get_template(
                self.route_info.meta['view'].lstrip('/'))
            local_tags = tags.create_builtin_tags(self.pod, doc=None)
            # NOTE: This should be done using get_template(... globals=...)
            # or passed as an argument into render but
            # it is not available included inside macros???
            # See: https://github.com/pallets/jinja/issues/688
            template.globals['g'] = local_tags

            try:
                serving_path = '/{}.html'.format(self.route_info.meta['key'])
                rendered_doc = rendered_document.RenderedDocument(
                    serving_path, template.render({
                        'doc': None,
                        'env': self.pod.env,
                        'podspec': self.pod.podspec,
                    }).lstrip())
                timer.stop_timer()
                return rendered_doc
            except Exception as err:
                text = 'Error building {}: {}'
                if self.pod:
                    self.pod.logger.exception(text.format(self, err))
                exception = errors.BuildError(text.format(self, err))
                exception.traceback = sys.exc_info()[2]
                exception.controller = self
                exception.exception = err
                raise exception


class RenderSitemapController(RenderController):
    """Controller for handling rendering for sitemaps."""

    @property
    def mimetype(self):
        """Determine headers to serve for https requests."""
        return mimetypes.guess_type(self.serving_path)[0]

    def load(self, source_dir):
        """Load the pod content from file system."""
        timer = self.pod.profile.timer(
            'RenderSitemapController.load',
            label='{}'.format(self.serving_path),
            meta=self.route_info.meta,
        ).start_timer()

        source_dir = self.clean_source_dir(source_dir)

        # Validate the path with the config filters.
        self.validate_path()

        try:
            rendered_path = '{}{}'.format(source_dir, self.serving_path)
            rendered_content = self.pod.storage.read(rendered_path)
            rendered_doc = rendered_document.RenderedDocument(
                self.serving_path, rendered_content)
            timer.stop_timer()
            return rendered_doc
        except Exception as err:
            text = 'Error building {}: {}'
            if self.pod:
                self.pod.logger.exception(text.format(self, err))
            exception = errors.BuildError(text.format(self, err))
            exception.traceback = sys.exc_info()[2]
            exception.controller = self
            exception.exception = err
            raise exception

    def render(self, jinja_env=None, request=None):
        """Render the document using the render pool."""
        timer = self.pod.profile.timer(
            'RenderSitemapController.render',
            label='{}'.format(self.serving_path),
            meta=self.route_info.meta,
        ).start_timer()

        # Validate the path with the config filters.
        self.validate_path()

        # Duplicate the routes to use the filters without messing up routing.
        temp_router = self.pod.router.__class__(self.pod)
        temp_router.add_all()

        # Sitemaps only show documents...?
        temp_router.filter('whitelist', kinds=['doc'])

        for sitemap_filter in self.route_info.meta.get('filters') or []:
            temp_router.filter(
                sitemap_filter['type'], collection_paths=sitemap_filter.get('collections'),
                paths=sitemap_filter.get('paths'), locales=sitemap_filter.get('locales'))

        # Need a custom root for rendering sitemap.
        root = os.path.join(utils.get_grow_dir(), 'pods', 'templates')
        jinja_env = self.pod.render_pool.custom_jinja_env(root=root)

        with jinja_env['lock']:
            if self.route_info.meta.get('template'):
                content = self.pod.read_file(self.route_info.meta['template'])
                template = jinja_env['env'].from_string(content)
            else:
                template = jinja_env['env'].get_template('sitemap.xml')

            try:
                docs = []
                for _, value, _ in temp_router.routes.nodes:
                    docs.append(self.pod.get_doc(value.meta['pod_path'], locale=value.meta['locale']))
                rendered_doc = rendered_document.RenderedDocument(
                    self.serving_path, template.render({
                        'pod': self.pod,
                        'env': self.pod.env,
                        'docs': docs,
                        'podspec': self.pod.podspec,
                    }).lstrip())
                timer.stop_timer()
                return rendered_doc
            except Exception as err:
                text = 'Error building {}: {}'
                if self.pod:
                    self.pod.logger.exception(text.format(self, err))
                exception = errors.BuildError(text.format(self, err))
                exception.traceback = sys.exc_info()[2]
                exception.controller = self
                exception.exception = err
                raise exception


class RenderStaticDocumentController(RenderController):
    """Controller for handling rendering for static documents."""

    def __init__(self, pod, serving_path, route_info, params=None, is_threaded=False):
        super(RenderStaticDocumentController, self).__init__(
            pod, serving_path, route_info, params=params, is_threaded=is_threaded)
        self._static_doc = None
        self._pod_path = None

    def __repr__(self):
        return '<RenderStaticDocumentController({})>'.format(self.route_info.meta['pod_path'])

    @property
    def pod_path(self):
        """Static doc for the controller."""
        if self._pod_path:
            return self._pod_path

        locale = self.route_info.meta.get(
            'locale', self.params.get('locale'))
        if 'pod_path' in self.route_info.meta:
            self._pod_path = self.route_info.meta['pod_path']
        else:
            for source_format in self.route_info.meta['source_formats']:
                path_format = '{}{}'.format(source_format, self.params['*'])
                self._pod_path = self.pod.path_format.format_static(
                    path_format, locale=locale)
                # Strip the fingerprint to get to the raw static file.
                self._pod_path = static_document.StaticDocument.strip_fingerprint(
                    self._pod_path)
                try:
                    # Throws an error when the document doesn't exist.
                    _ = self.pod.get_static(self._pod_path, locale=locale)
                    break
                except errors.DocumentDoesNotExistError:
                    self._pod_path = None
        return self._pod_path

    @property
    def static_doc(self):
        """Static doc for the controller."""
        if not self._static_doc:
            locale = self.route_info.meta.get(
                'locale', self.params.get('locale'))
            self._static_doc = self.pod.get_static(self.pod_path, locale=locale)
        return self._static_doc

    @property
    def mimetype(self):
        """Determine headers to serve for https requests."""
        return mimetypes.guess_type(self.serving_path)[0]

    def get_http_headers(self):
        """Determine headers to serve for http requests."""
        headers = super(RenderStaticDocumentController, self).get_http_headers()
        if self.pod_path is None:
            return headers
        path = self.pod.abs_path(self.static_doc.pod_path)
        self.pod.storage.update_headers(headers, path)
        modified = self.pod.storage.modified(path)
        time_obj = datetime.datetime.fromtimestamp(modified).timetuple()
        time_format = '%a, %d %b %Y %H:%M:%S GMT'
        headers['Last-Modified'] = time.strftime(time_format, time_obj)
        headers['ETag'] = '"{}"'.format(headers['Last-Modified'])
        headers['X-Grow-Pod-Path'] = self.static_doc.pod_path
        if self.static_doc.locale:
            headers['X-Grow-Locale'] = self.static_doc.locale
        return headers

    def load(self, source_dir):
        """Load the pod content from file system."""
        timer = self.pod.profile.timer(
            'RenderStaticDocumentController.load', label=self.serving_path,
            meta={'path': self.serving_path}).start_timer()

        source_dir = self.clean_source_dir(source_dir)

        # Validate the path with the static config specific filter.
        self.validate_path(self.route_info.meta['path_filter'])

        rendered_path = '{}{}'.format(source_dir, self.serving_path)
        rendered_content = self.pod.storage.read(rendered_path)
        rendered_doc = rendered_document.RenderedDocument(
            self.serving_path, rendered_content)
        timer.stop_timer()
        return rendered_doc

    def render(self, jinja_env=None, request=None):
        """Read the static file."""
        timer = self.pod.profile.timer(
            'RenderStaticDocumentController.render', label=self.serving_path,
            meta={'path': self.serving_path}).start_timer()

        if not self.pod_path or not self.pod.file_exists(self.pod_path):
            text = '{} was not found in static files.'
            raise errors.RouteNotFoundError(text.format(self.serving_path))

        # Validate the path with the static config specific filter.
        self.validate_path(self.route_info.meta['path_filter'])

        rendered_content = self.pod.read_file(self.pod_path)
        rendered_content = self.pod.extensions_controller.trigger(
            'post_render', self.static_doc, rendered_content)
        rendered_doc = rendered_document.RenderedDocument(
            self.serving_path, rendered_content)
        timer.stop_timer()
        return rendered_doc



"""Routes trie for mapping grow documents to paths."""

import collections
from grow.common import utils


PREFIX_PARAMETER = ':'
PREFIX_WILDCARD = '*'
URL_SEPARATOR = '/'
SHARD_KEY_DEFAULT = '_default'
TEMPLATE_CHAR = '{'


class Error(Exception):
    """Base routes error."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class MissingOptionError(Error):
    """Error when there is a missing option when using templated path."""

    def __init__(self, path):
        super(MissingOptionError, self).__init__(
            'Missing a templated option in path: {}'.format(path))
        self.path = path


class PathConflictError(Error):
    """Error when there is a conflict of paths in the routes trie."""

    def __init__(self, path, value, existing):
        super(PathConflictError, self).__init__(
            'Path already exists: {} ({} != {})'.format(path, existing, value))
        self.path = path
        self.value = value
        self.existing = existing


class PathParamNameConflictError(Error):
    """Error when there is a conflict of path param names in the routes trie."""

    def __init__(self, path, new_param, existing_param):
        super(PathParamNameConflictError, self).__init__(
            'Path param differs from existing param: {} ({} vs {})'.format(
                path, new_param, existing_param))
        self.path = path
        self.new_param = new_param
        self.existing_param = existing_param


class Routes(object):
    """Routes container for mapping paths to documents."""

    def __add__(self, other):
        """Allow adding routes togethers to form a new routes."""
        routes = self.__class__()
        # Add all the nodes from the current Routes.
        for path, value, options in self.nodes:
            routes.add(path, value, options=options)
        # Add all the nodes from the other Routes.
        for path, value, options in other.nodes:
            routes.add(path, value, options=options)
        return routes

    def __init__(self):
        self._root = RouteTrie()

    def __len__(self):
        return len(self._root)

    @property
    def nodes(self):
        """Generator for returning all the nodes in the routes."""
        for item in self._root.nodes:
            yield item

    @property
    def paths(self):
        """Generator for returning all the paths in the routes."""
        for path, _, _ in self._root.nodes:
            yield path

    def add(self, path, value, options=None):
        """Adds a document to the routes trie."""
        if not path:
            return
        self._root.add(path, value, options=options)

    def export(self):
        """Export the routes into a serializable format."""
        routing_info = {}
        for path, value, options in self.nodes:
            value = value.export() if hasattr(value, 'export') else value
            routing_info[path] = {
                'value': value,
                'options': options,
            }
        return routing_info

    def filter(self, func):
        """Filters out the nodes that do not match the filter."""
        if not func:
            return
        return self._root.filter(func)

    def match(self, path):
        """Uses a path to attempt to match a path in the routes."""
        return self._root.match(path)

    def reset(self):
        """Resets the routes."""
        self._root = RouteTrie()

    def remove(self, path):
        """Removes a path from the routes."""
        return self._root.remove(path)

    def shard(self, shard_count, current_shard, attr='kind'):
        """Removes paths from the routes based on sharding rules."""

        if shard_count <= 1:
            raise ValueError('Shard count needs to be greater than 1')

        if shard_count >= 50:
            raise ValueError('Shard count needs to be less than 50')

        if shard_count < current_shard:
            raise ValueError('Shard count needs to be larger than the current shard')

        if current_shard < 1:
            raise ValueError('Current shard needs to be at least 1')

        return self._root.shard(shard_count, current_shard, attr=attr)

    def update(self, other):
        """Allow updating the current routes with other Routes."""
        # Add all the nodes from the other Routes.
        for path, value, options in other.nodes:
            self.add(path, value, options=options)


class RoutesSimple(Routes):
    """Routes container for mapping simple paths to values."""

    def __init__(self):
        super(RoutesSimple, self).__init__()
        self._root = RoutesDict()

    def reset(self):
        """Resets the routes."""
        self._root = RoutesDict()


class RoutesDict(object):
    """Dictionary based routing tree for faster creation."""

    def __init__(self):
        self._root = {}

    def __len__(self):
        return len(self._root)

    @property
    def nodes(self):
        """Generator for returning all nodes in the trie."""
        for path in sorted(self._root):
            yield path, self._root[path], None

    # pylint: disable=unused-argument
    def add(self, path, value, options=None):
        """Add a new doc to the route trie."""
        if path in self._root and self._root[path] != value:
            raise PathConflictError(path, value, self._root[path])
        self._root[path] = value

    def filter(self, func):
        """Filters out the nodes that do not match the filter."""
        count = 0
        remove_paths = []
        for path in self._root:
            if not func(path, self._root[path]):
                remove_paths.append(path)

        for path in remove_paths:
            self._root.pop(path, None)
            count = count + 1

        return count

    def match(self, path):
        """Matches a path against the known routes looking for a match."""
        value = self._root.get(path, None)
        if value is None:
            return None
        return MatchResult(path, value)

    def remove(self, path):
        """Removes a path from the trie."""
        value = self._root.pop(path, None)
        if value is None:
            return None
        return MatchResult(path, value)

    def shard(self, shard_count, current_shard, attr='kind'):
        """Removes paths from the routes based on sharding rules."""
        shard_index = current_shard - 1
        counters = {}
        remove_paths = []
        for path in sorted(self._root):
            counter_key = SHARD_KEY_DEFAULT

            # Use the attribute as a counter to equally distribute routes based
            # on an attribute in the value.
            if attr:
                value = self._root.get(path, None)
                if value is not None:
                    counter_key = getattr(value, attr, SHARD_KEY_DEFAULT)

            count = counters.get(counter_key, 0)

            # Remove all paths that do not match the current shard.
            if count % shard_count != shard_index:
                remove_paths.append(path)

            counters[counter_key] = count + 1

        for path in remove_paths:
            self._root.pop(path, None)


class RouteTrie(object):
    """A trie for routes."""

    def __init__(self):
        self._root = RouteNode()

    def __len__(self):
        i = 0
        for _ in self.nodes:
            i = i + 1
        return i

    @staticmethod
    def clean_path(path):
        """Clean the path to work with the trie."""
        if path and path[0] == URL_SEPARATOR:
            path = path[1:]
        if path and path[-1] == URL_SEPARATOR:
            path = path[0:-1]
        return path

    @staticmethod
    def segments(path):
        """Clean the and break the path to work with the trie."""
        path = RouteTrie.clean_path(path)
        return collections.deque(path.split(URL_SEPARATOR))

    @property
    def nodes(self):
        """Generator for returning all nodes in the trie."""
        for item in self._root.nodes:
            yield item

    def add(self, path, value, options=None):
        """Add a new doc to the route trie."""
        segments = self.segments(path)
        self._root.add(segments, path, value, options=options)

    def filter(self, func):
        """Filters out the nodes that do not match the filter."""
        return self._root.filter(func)

    def match(self, path):
        """Matches a path against the known trie looking for a match."""
        segments = self.segments(path)
        return self._root.match(segments)

    def remove(self, path):
        """Removes a path from the trie."""
        segments = self.segments(path)
        return self._root.remove(segments)

    # pylint: disable=unused-argument
    def shard(self, shard_count, current_shard, attr='kind'):
        """Removes paths from the routes based on sharding rules."""
        # Sharding doesn't work on the routing trie since it uses patterns
        # and would not equally distribute the routes.
        raise NotImplementedError('Sharding cannot be done on routing trie.')


class RouteNode(object):
    """Individual node in the routes trie."""

    def __init__(self, param_name=None):
        super(RouteNode, self).__init__()
        self.path = None
        self.paths = set()
        self.value = None
        self.options = None
        self.param_name = param_name
        self.param_options = None
        self._dynamic_children = {}
        self._static_children = {}

    def __repr__(self):
        return "<RouteNode({}, {})>".format(self.path, self.param_name)

    @property
    def nodes(self):
        """Generator for walking through the node nodes.

        Yields:
            Path, value at this node and all children.
        """
        if self.path is not None:
            yield self.path, self.value, self.options

        # Yield nodes in the path order.
        for key in sorted(self._static_children):
            for item in self._static_children[key].nodes:
                yield item

        for key in sorted(self._dynamic_children):
            for item in self._dynamic_children[key].nodes:
                yield item

    @property
    def is_templated(self):
        """Does the node have a templated path?"""
        return len(self.paths) > 1

    def _dynamic_params(self, matched, last_segment):
        # Parameterized nodes need to add in their param when returning.
        if matched is not None and self.param_name:
            matched.params[self.param_name] = last_segment
        return matched

    @staticmethod
    def _dynamic_paths(path, options):
        """Use the path to format using the options to determine all paths."""
        path = RouteTrie.clean_path(path)
        possible_paths = set([path])

        # Check for string formatting.
        if not options or '{' not in path:
            return possible_paths

        for key in options:
            for option in options[key]:
                format_keys = {key: option}
                new_paths = []
                for possible_path in possible_paths:
                    new_paths.append(utils.safe_format(possible_path, **format_keys))
                possible_paths = possible_paths.union(new_paths)

        paths = set()

        # Don't include paths that still have format symbols.
        for possible_path in possible_paths:
            if '{' not in possible_path:
                paths.add(possible_path)

        if not paths:
            raise MissingOptionError(path)

        return paths

    def add(self, segments, path, value, options=None):
        """Recursively add into the trie based upon the given segments."""

        if not segments:
            if self.path and self.value != value:
                raise PathConflictError(path, value, self.value)
            self.path = path
            self.value = value
            self.options = options
            # Generate all possible values for the path using the options.
            self.paths = self._dynamic_paths(path, options)
            return

        segment = segments.popleft()

        # Insert as a parameterized node.
        if segment and segment.startswith(PREFIX_PARAMETER):
            segment = segment[1:]  # Don't need the prefix character.
            if PREFIX_PARAMETER in self._dynamic_children:
                # If there is already a prefix in use, cannot have a conflicting
                # param_name or it causes issues with params on match.
                if self._dynamic_children[PREFIX_PARAMETER].param_name != segment:
                    raise PathParamNameConflictError(
                        path, segment, self._dynamic_children[PREFIX_PARAMETER].param_name)

                # If there are options that match the parameter, add them in.
                if options and segment in options:
                    self._dynamic_children[PREFIX_PARAMETER].add_param_options(
                        options[segment])
            if PREFIX_PARAMETER not in self._dynamic_children:
                new_node = RouteNode(param_name=segment)
                if options and segment in options:
                    new_node.add_param_options(options[segment])
                self._dynamic_children[PREFIX_PARAMETER] = new_node
            self._dynamic_children[PREFIX_PARAMETER].add(
                segments, path, value, options=options)
            return

        # Insert as a wildcard node.
        if segment and segment.startswith(PREFIX_WILDCARD):
            segment = segment[1:]  # Don't need the prefix character.
            if (PREFIX_WILDCARD in self._dynamic_children
                    and self._dynamic_children[PREFIX_WILDCARD].value != value):
                raise PathConflictError(
                    path, value, self._dynamic_children[PREFIX_WILDCARD].value)
            new_node = RouteWildcardNode(param_name=segment or PREFIX_WILDCARD)
            new_node.add([], path, value, options=options)
            self._dynamic_children[PREFIX_WILDCARD] = new_node
            return

        # Add a static node.
        if segment not in self._static_children:
            self._static_children[segment] = RouteNode()
        self._static_children[segment].add(
            segments, path, value, options=options)

    def add_param_options(self, options):
        """Add options for parameter."""
        if not self.param_options:
            self.param_options = set()
        self.param_options |= set(options)

    def filter(self, func):
        """Filters out the nodes that do not match the filter."""
        count = 0
        if self.path is not None:
            if not func(self.path, self.value):
                count = count + 1
                self.path = None
                self.value = None

        for key in self._static_children:
            count = count + self._static_children[key].filter(func)

        for key in self._dynamic_children:
            count = count + self._dynamic_children[key].filter(func)

        return count

    # pylint: disable=too-many-return-statements
    def match(self, segments, last_segment=None):
        """Performs the trie matching to find a doc in the trie."""

        # If this is a parameterized node and it needs to match the options.
        if last_segment and self.param_options:
            if last_segment not in self.param_options:
                return None

        if not segments:
            # Check for removed nodes.
            if self.path is None:
                return None

            # Check for templated paths to have a direct match with paths.
            if self.is_templated and last_segment not in self.paths:
                return None

            matched = MatchResult(self.path, self.value)
            return self._dynamic_params(matched, last_segment)

        segment = segments.popleft()

        # Match static children first, then fallback to dynamic.
        if segment in self._static_children:
            matched = self._static_children[segment].match(
                segments, last_segment=segment)
            if matched is not None:
                return self._dynamic_params(matched, last_segment)
        else:
            # Check for templated static children.
            for key in self._static_children:
                child = self._static_children[key]
                if child.is_templated:
                    matched = child.match(segments, last_segment=segment)
                    if matched is not None:
                        return self._dynamic_params(matched, last_segment)

        # Check if this is a parameterized segement.
        if PREFIX_PARAMETER in self._dynamic_children:
            matched = self._dynamic_children[PREFIX_PARAMETER].match(
                segments, last_segment=segment)
            if matched is not None:
                return self._dynamic_params(matched, last_segment)

        # Check for the wildcard catch all.
        if PREFIX_WILDCARD in self._dynamic_children:
            matched = self._dynamic_children[PREFIX_WILDCARD].match(
                segments, last_segment=segment)
            return self._dynamic_params(matched, last_segment)

        # Add the current segment back since this is not a match.
        segments.appendleft(segment)
        return None

    def remove(self, segments):
        """Finds and removes a path in the trie."""

        if not segments:
            removed = MatchResult(self.path, self.value)
            self.path = None
            self.value = None
            return removed
        segment = segments.popleft()

        # Static nodes.
        if segment in self._static_children:
            return self._static_children[segment].remove(segments)

        # Parameter nodes.
        if segment and segment[0] == PREFIX_PARAMETER:
            if PREFIX_PARAMETER in self._dynamic_children:
                return self._dynamic_children[PREFIX_PARAMETER].remove(segments)

        # Wildcard nodes.
        if segment and segment[0] == PREFIX_WILDCARD:
            if PREFIX_WILDCARD in self._dynamic_children:
                return self._dynamic_children[PREFIX_WILDCARD].remove(segments)

        return None


class RouteWildcardNode(RouteNode):
    """Individual wildcard node in the routes trie."""

    def match(self, segments, last_segment=None):
        """Matches the rest of the segments as the wildcard portion."""

        # Handle deleted node.
        if self.path is None:
            return None

        # Add the last segment back to be joined.
        if last_segment:
            segments.appendleft(last_segment)
        segment = URL_SEPARATOR.join(segments)

        return self._dynamic_params(
            MatchResult(self.path, self.value), segment)

    def remove(self, segments):
        """Finds and removes a path in the trie."""
        removed = MatchResult(self.path, self.value)
        self.path = None
        self.value = None
        return removed


# pylint: disable=too-few-public-methods
class MatchResult(object):
    """Node information for a trie match."""

    def __init__(self, path, value, params=None):
        self.path = path
        self.value = value
        self.params = params if params is not None else {}



"""Git deploy destination."""

import logging
import os
import re
import shutil
import subprocess
import tempfile
import git
from protorpc import messages
from grow.common import utils as common_utils
from grow.deployments.destinations import base
from grow.pods import env
from grow.routing import router
from grow.storage import file_storage


ONLINE_REPO_REGEX = \
    r'((git|ssh|http(s)?)|(git@[\w\.]+))(:(//)?)([\w\.@\:/\-~]+)(\.git)(/)?'


class Config(messages.Message):
    env = messages.MessageField(env.EnvConfig, 1)
    repo = messages.StringField(2)
    branch = messages.StringField(3, default='master')
    root_dir = messages.StringField(4, default='')
    keep_control_dir = messages.BooleanField(5, default=False)
    filters = messages.MessageField(router.FilterConfig, 6, repeated=True)
    tags = messages.StringField(7, repeated=True)


class GitDestination(base.BaseDestination):
    """Git deploy destination."""
    KIND = 'git'
    Config = Config
    storage = file_storage.FileStorage

    def __init__(self, *args, **kwargs):
        super(GitDestination, self).__init__(*args, **kwargs)
        self.adds = set()
        self.deletes = set()
        self._original_branch_name = None

    def __str__(self):
        if self.is_remote:
            prefix = self.config.repo
        else:
            prefix = 'git:{}'.format(self.config.repo)
        return '{}@{}:{}'.format(prefix, self.config.branch, self.config.root_dir)

    @common_utils.cached_property
    def is_remote(self):
        return re.match(ONLINE_REPO_REGEX, self.config.repo)

    @common_utils.cached_property
    def repo_path(self):
        if self.is_remote:
            return tempfile.mkdtemp()
        else:
            return os.path.abspath(self.config.repo)

    @common_utils.cached_property
    def repo(self):
        if self.is_remote:
            return git.Repo.init(self.repo_path)
        return git.Repo(self.repo_path)

    def _checkout(self, branch=None):
        branch = branch or self.config.branch
        try:
            self.repo.git.checkout(b=branch)
        except git.GitCommandError as e:
            if e.status == 128:
                self.repo.git.checkout(branch)

    def prelaunch(self, dry_run=False):
        self._original_branch_name = self.repo.active_branch.name
        self._checkout()
        if self.is_remote:
            self.remote = git.remote.Remote.add(self.repo, 'origin', self.config.repo)
            try:
                logging.info('Pulling from {}...'.format(self.config.branch))
                self.repo.git.pull('origin', self.config.branch)
            except git.GitCommandError as e:
                # Pass on this error, which will create a new branch upon pushing.
                if "Couldn't find remote ref" not in e.stderr:
                    raise

    def create_commit_message(self):
        editor = os.getenv('EDITOR')
        commit_message_path = os.path.join(self.repo.git_dir, 'COMMIT_EDITMSG')
        fp = open(commit_message_path, 'w')
        fp.write(self._diff.what_changed or '')
        fp.close()
        if editor and self._confirm:
            subprocess.call('{} {}', editor, commit_message_path, shell=True)
        content = open(commit_message_path).read()
        return content

    def postlaunch(self, dry_run=False):
        if dry_run:
            if self.is_remote:
                shutil.rmtree(self.repo_path)
            return

        if self.adds:
            self.repo.index.add(self.adds)
        if self.deletes:
            self.repo.index.remove(self.deletes)

        if not self.adds and not self.deletes:
            logging.info('No changes, aborting.')
            return

        # TODO: Replace with a user-configurable launch message.
        self.repo.index.commit(self.create_commit_message())

        if self.is_remote:
            logging.info('Pushing to origin...')
            self.repo.git.push('origin', self.config.branch)
            shutil.rmtree(self.repo_path)
        elif self._original_branch_name is not None:
            self._checkout(self._original_branch_name)

    def read_file(self, path):
        path = os.path.join(self.repo_path, self.config.root_dir.lstrip('/'),
                            path.lstrip('/'))
        return self.storage.read(path)

    def delete_control_file(self, path):
        path = os.path.join(self.control_dir, path.lstrip('/'))
        out_path = self.delete_file(path)
        # Control files should remain in the index, for now.
        self.deletes.remove(out_path)

    def delete_file(self, path):
        out_path = os.path.join(self.repo_path, self.config.root_dir.lstrip('/'),
                                path.lstrip('/'))
        self.storage.delete(out_path)
        self.deletes.add(out_path)
        if out_path in self.adds:
            self.adds.remove(out_path)
        return out_path

    def write_file(self, rendered_doc):
        path = rendered_doc.path
        content = rendered_doc.read()
        out_path = os.path.join(self.repo_path, self.config.root_dir.lstrip('/'),
                                path.lstrip('/'))
        self.storage.write(out_path, content)
        self.adds.add(out_path)
        if out_path in self.deletes:
            self.deletes.remove(out_path)



"""Custom json encoder for handling Grow classes."""

import datetime
import json
from grow.documents import document
from grow.documents import document_fields


class GrowJSONEncoder(json.JSONEncoder):
    """Encoder for serializing Grow objects."""

    # pylint: disable=method-hidden
    def default(self, o):
        """Attempt to encode known objects."""
        if isinstance(o, datetime.date):
            return o.isoformat()

        if isinstance(o, datetime.datetime):
            return o.isoformat()

        if isinstance(o, document.Document):
            # pylint: disable=protected-access
            return o.fields._data

        if isinstance(o, document_fields.DocumentFields):
            # pylint: disable=protected-access
            return o._data

        if isinstance(o, set):
            return list(o)

        return json.JSONEncoder.default(self, o)



"""Translation catalogs."""

from datetime import datetime
import logging
import os
import re
import textwrap
import fnmatch
from babel import util
from babel.messages import catalog
from babel.messages import mofile
from babel.messages import pofile
from babel.util import odict
from grow.pods import messages


class Catalog(catalog.Catalog):

    def __init__(self, basename='messages.po', locale=None, pod=None,
                 dir_path=None):
        dir_path = 'translations' if dir_path is None else dir_path
        self.locale = locale
        self.pod = pod
        if self.locale is None:
            self.pod_path = os.path.join(dir_path, basename)
            is_template = True
        else:
            self.pod_path = os.path.join(dir_path, str(locale), 'LC_MESSAGES',
                                         basename)
            is_template = False
        super(Catalog, self).__init__(locale=self.locale)
        if not is_template and self.exists:
            self.load()

    def __repr__(self):
        return '<Catalog: {}>'.format(self.pod_path)

    @property
    def mimetype(self):
        return 'text/x-gettext-translation'

    def load(self, pod_path=None):
        # Use the "pod_path" argument to load another catalog (such as a template
        # catalog) into this one.
        if pod_path is None:
            pod_path = self.pod_path
        if not self.pod.file_exists(pod_path):
            self.pod.write_file(pod_path, '')
        po_file = self.pod.open_file(pod_path)
        try:
            try:
                babel_catalog = pofile.read_po(po_file, self.locale)
            except:
                self.pod.logger.error('Error parsing catalog for: {}'.format(self.locale))
                raise
        finally:
            po_file.close()
        attr_names = [
            '_messages',
            '_num_plurals',
            '_plural_expr',
            'charset',
            'copyright_holder',
            'creation_date',
            'domain',
            'fuzzy',
            'language_team',
            'last_translator',
            'locale',
            'msgid_bugs_address',
            'obsolete',
            'project',
            'revision_date',
            'version',
        ]
        for name in attr_names:
            setattr(self, name, getattr(babel_catalog, name))

    @property
    def exists(self):
        return self.pod.file_exists(self.pod_path)

    @property
    def content(self):
        return self.pod.read_file(self.pod_path)

    def to_message(self):
        catalog_message = messages.CatalogMessage()
        catalog_message.messages = []
        for message in self:
            message_message = messages.MessageMessage()
            message_message.msgid = message.id
            message_message.msgstr = message.string
            catalog_message.messages.append(message_message)
        return catalog_message

    def save(self, include_header=False):
        for message in self:
            message.flags.discard('python-format')
        if not self.pod.file_exists(self.pod_path):
            self.pod.write_file(self.pod_path, '')
        outfile = self.pod.open_file(self.pod_path, mode='wb')
        Catalog.set_header_comment(self.pod, self)
        pofile.write_po(
            outfile,
            self,
            omit_header=(not include_header),
            sort_output=True,
            sort_by_file=True,
            include_previous=True,
            width=80)
        outfile.close()

    def init(self, template_path, include_header=False):
        self.load(template_path)
        self.revision_date = datetime.now(util.LOCALTZ)
        self.fuzzy = False
        self.save(include_header=include_header)

    def update_using_catalog(self, catalog_to_merge, use_fuzzy_matching=False,
                             include_obsolete=False):
        super(Catalog, self).update(
            catalog_to_merge, no_fuzzy_matching=(not use_fuzzy_matching))
        # Don't use gettext's obsolete functionality as it polutes files: merge
        # into main translations if anything
        if include_obsolete:
            self.merge_obsolete()
        self.obsolete = odict()

    def update(self, template_path=None, use_fuzzy_matching=False,
               include_obsolete=False, include_header=False):
        """Updates catalog with messages from a template."""
        if template_path is None:
            template_path = os.path.join('translations', 'messages.pot')
        if not self.exists:
            self.init(template_path)
            return
        template_file = self.pod.open_file(template_path)
        template = pofile.read_po(template_file)
        super(Catalog, self).update(
            template, no_fuzzy_matching=(not use_fuzzy_matching))
        # Don't use gettext's obsolete functionality as it polutes files: merge
        # into main translations if anything
        if include_obsolete:
            self.merge_obsolete()
        self.obsolete = odict()
        self.save(include_header=include_header)

    def merge_obsolete(self):
        """Copy obsolete terms into the main catalog."""
        for msgid, message in self.obsolete.items():
            self[msgid] = message

    @property
    def mo_path(self):
        mo_dirpath = os.path.dirname(self.pod_path)
        return os.path.join(mo_dirpath, 'messages.mo')

    @property
    def mo_modified(self):
        try:
            return self.pod.file_modified(self.mo_path)
        except OSError:
            return None

    @property
    def modified(self):
        try:
            return self.pod.file_modified(self.pod_path)
        except OSError:
            return None

    @property
    def needs_compilation(self):
        if self.mo_modified is None or self.modified is None:
            return True
        return self.modified > self.mo_modified

    def _skip_compile_error(self, error):
        # Reduces logspam by hiding errors related to placeholders.
        if ('incompatible format for placeholder' in str(error)
                 or 'placeholders are incompatible' in str(error)):
            return True
        return False

    def compile(self):
        self.pod.catalogs.clear_gettext_cache()
        localization = self.pod.podspec.localization
        if localization is None:
            return
        compile_fuzzy = localization.get('compile_fuzzy')
        mo_filename = self.mo_path
        num_translated = 0
        num_total = 0
        for message in list(self)[1:]:
            if message.string:
                num_translated += 1
            num_total += 1
        try:
            for message, errors in self.check():
                for error in errors:
                    if self._skip_compile_error(error):
                        continue
                    text = 'Error compiling ({}:{}): {}'
                    message = text.format(self.locale, message.lineno, error)
                    self.pod.logger.error(message)
        except IOError:
            self.pod.logger.info('Skipped catalog check for: {}'.format(self))
        text = 'Compiled: {} ({}/{})'
        self.pod.logger.info(text.format(self.locale, num_translated, num_total))
        mo_file = self.pod.open_file(mo_filename, 'wb')
        try:
            mofile.write_mo(mo_file, self, use_fuzzy=compile_fuzzy)
        finally:
            mo_file.close()

    @classmethod
    def _message_in_paths(cls, message, paths):
        location_paths = set([path for path, unused_lineno in message.locations])
        for path in paths:
            for location_path in location_paths:
                # Support pod paths and filesystem paths for tab completion.
                location_path = location_path.lstrip('/')
                path = path.lstrip('/')
                matched = fnmatch.fnmatch(location_path, path)
                if matched:
                    return True
        return False

    def list_untranslated(self, paths=None):
        """Returns untranslated messages, including fuzzy translations."""
        untranslated = []
        for message in self:
            if paths and not Catalog._message_in_paths(message, paths):
                continue
            # Ensure fuzzy messages have a message.id otherwise we'd include
            # the header as part of the results, which we don't want.
            if not message.string or (message.fuzzy and message.id):
                untranslated.append(message)
        return untranslated

    @staticmethod
    def set_header_comment(pod, catalog):
        project_title = (
            pod.yaml.get('translators', {})
            .get('project_title', 'Untitled Grow Website'))
        instructions = (
            pod.yaml.get('translators', {}).get('instructions'))
        comment = ''
        if catalog.locale:
            lang_display_name = catalog.locale.language
            comment += """
            PROJECT TITLE:
            {} ({})
            """.format(project_title, lang_display_name)
        else:
            comment += """
            PROJECT TITLE:
            {}
            """.format(project_title)
        comment = textwrap.dedent(comment)
        if instructions:
            instructions = textwrap.dedent(instructions)
            comment += """
            PROJECT INSTRUCTIONS:
            {}
            """.format(instructions)
        comment += ''
        comment = ['# {}'.format(line.lstrip()) for line in comment.split('\n')]
        comment = comment[1:]  # Strip the first blank line.
        catalog.header_comment = '\n'.join(comment)



"""Command for downloading translations from remote source."""

import os
import click
from grow.commands import shared
from grow.common import rc_config
from grow.pods import pods
from grow import storage


CFG = rc_config.RC_CONFIG.prefixed('grow.translations.download')


@click.command(name='download')
@shared.pod_path_argument
@shared.locale_option(help_text='Which locale(s) to download. If unspecified,'
                                ' translations for all locales will be downloaded.')
@shared.service_option
@shared.include_obsolete_option(CFG)
def translations_download(pod_path, locale, service, include_obsolete):
    """Downloads translations from a translation service."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    pod = pods.Pod(root, storage=storage.FileStorage)
    with pod.profile.timer('grow_translations_download'):
        translator = pod.get_translator(service)
        translator.download(locales=locale, include_obsolete=include_obsolete)
    return pod



"""Bulk error handling and reporting."""

import logging
import traceback

class BulkErrors(Exception):
    """Base bulk errors."""

    def __init__(self, message, errors):
        super(BulkErrors, self).__init__(message)
        self.message = message
        self.errors = errors

def display_bulk_errors(bulk_errors):
    """Nicely log and show the bulk errors."""
    for error in bulk_errors.errors:
        try:
            logging.error(error.message)
        except AttributeError:
            logging.error(error)
        try:
            logging.error(error.err.message)
        except AttributeError:
            logging.error(error.err)
        traceback.print_tb(error.err_tb)
        logging.info('')
    logging.error(bulk_errors.message)



"""Translation catalog container."""

import io
import collections
import gettext
import os
import tokenize
import texttable
from babel import support
from babel.messages import catalog as babel_catalog
from babel.messages import extract
from babel.messages import mofile
from babel.messages import pofile
import click
from grow.common import utils
from grow.pods import messages
from grow.translations import catalogs
from grow.translations import importers
from grow.translations import locales as grow_locales


_IGNORED_PREFIXS = (
    '.',
    '_',
)
_TRANSLATABLE_EXTENSIONS = (
    '.html',
    '.md',
    '.yaml',
    '.yml',
)


class Error(Exception):

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class UsageError(Error, click.UsageError):
    pass


class Catalogs(object):
    root = '/translations'

    def __init__(self, pod, template_path=None):
        self.pod = pod
        self._gettext_translations = {}
        if template_path:
            self.template_path = os.path.expanduser(template_path)
        else:
            self.template_path = os.path.join(Catalogs.root, 'messages.pot')
        self.root = os.path.dirname(self.template_path)

    def __repr__(self):
        return '<Catalogs: {}>'.format(self.template_path)

    def diff(self, other_catalogs, out_dir):
        """Produces a diff between this directory of catalogs, and another set
        of catalogs, writing the results to a directory."""
        diffed_locales_to_catalogs = collections.defaultdict(int)
        if not out_dir.endswith('messages.pot'):
            out_dir = os.path.join(out_dir, 'messages.pot')
        diffed_catalogs = self.pod.get_catalogs(out_dir)
        for this_catalog in self:
            locale = this_catalog.locale
            other_catalog = other_catalogs.get(locale, dir_path=other_catalogs.root)
            diffed_catalog = diffed_catalogs.get(locale, dir_path=diffed_catalogs.root)
            for message in this_catalog:
                # Skip empty messages.
                if not message.id:
                    continue
                # Skip messages we have in the other catalog.
                if message.id in other_catalog:
                    continue
                diffed_catalog[message.id] = message
                diffed_locales_to_catalogs[locale] += 1
            diffed_catalog.save()
        for locale, num_diff in diffed_locales_to_catalogs.items():
            self.pod.logger.info('Found different messages for {} -> {}'.format(locale, num_diff))

    def get(self, locale, basename='messages.po', dir_path=None):
        return catalogs.Catalog(basename, locale, pod=self.pod, dir_path=dir_path)

    def get_template(self, basename='messages.pot'):
        template_catalog = catalogs.Catalog(basename, None, pod=self.pod)
        if template_catalog.exists:
            template_catalog.load()
        return template_catalog

    def list_locales(self):
        locales = set()
        for path in self.pod.list_dir(self.root):
            parts = path.split('/')
            if len(parts) > 2:
                locales.add(parts[1])
        return list(locales)

    def validate_locales(self, locales):
        for locale in locales:
            parsed_locale = grow_locales.Locale.parse(locale)
            if str(parsed_locale) != locale:
                text = 'WARNING: Translation directories are case sensitive (move {} -> {}).'
                self.pod.logger.warning(text.format(locale, parsed_locale))

    def __iter__(self):
        for locale in self.list_locales():
            yield self.get(locale)

    def __len__(self):
        return len([catalog for catalog in self])

    def compile(self, force=False):
        self.clear_gettext_cache()
        locales = self.list_locales()
        self.validate_locales(locales)
        skipped_locales = []
        for locale in locales:
            catalog = self.get(locale)
            if not catalog.exists:
                skipped_locales.append(locale)
                continue
            if force or catalog.needs_compilation:
                catalog.compile()
        if skipped_locales:
            skipped_locales.sort()
            text = 'No translations to compile -> {}'
            self.pod.logger.info(text.format(', '.join(skipped_locales)))

    def to_message(self):
        message = messages.CatalogsMessage()
        message.catalogs = []
        for locale in self.list_locales():
            catalog = self.get(locale)
            message.catalogs.append(catalog.to_message())
        return message

    def get_extract_config(self, include_obsolete=None, localized=None,
                           include_header=None, use_fuzzy_matching=None):
        extract_config = self.pod.yaml.get(
            'localization', {}).get('extract', {})
        if include_obsolete is None:
            include_obsolete = extract_config.get('include_obsolete', False)
        if localized is None:
            localized = extract_config.get('localized', False)
        if include_header is None:
            include_header = extract_config.get('include_header', False)
        if use_fuzzy_matching is None:
            use_fuzzy_matching = extract_config.get('fuzzy_matching', False)
        return include_obsolete, localized, include_header, use_fuzzy_matching

    def init(self, locales, include_header=None):
        _, _, include_header, _ = \
            self.get_extract_config(include_header=include_header)
        for locale in locales:
            catalog = self.get(locale)
            catalog.init(template_path=self.template_path,
                         include_header=include_header)

    def update(self, locales, use_fuzzy_matching=None, include_header=None,
               include_obsolete=None):
        include_obsolete, _, include_header, use_fuzzy_matching = \
            self.get_extract_config(include_header=include_header,
                    include_obsolete=include_obsolete,
                    use_fuzzy_matching=use_fuzzy_matching)
        for locale in locales:
            catalog = self.get(locale)
            self.pod.logger.info('Updating: {}'.format(locale))
            catalog.update(template_path=self.template_path,
                           use_fuzzy_matching=use_fuzzy_matching,
                           include_header=include_header,
                           include_obsolete=include_obsolete)

    def import_translations(self, path=None, locale=None, content=None,
                            include_obsolete=True, untranslated=False):
        importer = importers.Importer(self.pod,
                include_obsolete=include_obsolete, untranslated=untranslated)
        if path:
            return importer.import_path(path, locale=locale)
        if content:
            return importer.import_content(content=content, locale=locale)

    def _get_or_create_catalog(self, template_path):
        exists = True
        if not self.pod.file_exists(template_path):
            self.pod.write_file(template_path, '')
            exists = False
        catalog = pofile.read_po(self.pod.open_file(template_path))
        return catalog, exists

    def _add_message(self, catalog, message):
        lineno, string, comments, context = message
        flags = set()
        if string in catalog:
            existing_message = catalog.get(string)
            flags = existing_message.flags
        return catalog.add(string, None, auto_comments=comments, context=context,
                           flags=flags)

    def _should_extract_as_csv(self, given_paths, path):
        ext = os.path.splitext(path)[-1]
        if ext != '.csv':
            return False
        return given_paths is None or path in given_paths

    def _should_extract_as_babel(self, given_paths, path):
        if os.path.splitext(path)[-1] not in _TRANSLATABLE_EXTENSIONS:
            return False
        return not given_paths or path in given_paths

    @staticmethod
    def _starts_with_paths(paths, path):
        for prefix in paths:
            if path.startswith(prefix):
                return True
        return False

    def extract(self, include_obsolete=None, localized=None, paths=None,
                include_header=None, locales=None, use_fuzzy_matching=None,
                audit=False, out_path=None):
        include_obsolete, localized, include_header, use_fuzzy_matching, = \
            self.get_extract_config(include_header=include_header,
                                    include_obsolete=include_obsolete, localized=localized,
                                    use_fuzzy_matching=use_fuzzy_matching)

        env = self.pod.get_jinja_env()
        # {
        #    locale1: locale1_catalog,
        #    locale2: locale2_catalog,
        #    ...
        # }
        # This is built up as we extract
        localized_catalogs = {}
        untagged_strings = []
        unlocalized_catalog = catalogs.Catalog(pod=self.pod)  # for localized=False case

        comment_tags = [
            ':',
        ]
        options = {
            'extensions': ','.join(env.extensions.keys()),
            'silent': 'false',
        }

        def _add_to_catalog(message, locales):
            # Add to all relevant catalogs
            for locale in locales:
                if locale not in localized_catalogs:
                    # Start with a new catalog so we can track what's obsolete:
                    # we'll merge it with existing translations later.
                    # *NOT* setting `locale` kwarg here b/c that will load existing
                    # translations.
                    localized_catalogs[locale] = catalogs.Catalog(pod=self.pod)
                localized_catalogs[locale][message.id] = message
            unlocalized_catalog[message.id] = message

        def _handle_field(path, locales, msgid, key, node, parent_node=None):
            if (not key
                    or not isinstance(msgid, str)
                    or not isinstance(key, str)):
                return
            if not key.endswith('@'):
                if msgid:
                    untagged_strings.append((path, msgid))
                return
            # Support gettext "extracted comments" on tagged fields:
            #   field@: Message.
            #   field@#: Extracted comment for field@.
            auto_comments = []
            if isinstance(node, dict):
                auto_comment = node.get('{}#'.format(key))
                if auto_comment:
                    auto_comments.append(auto_comment)
            elif isinstance(node, list) and parent_node:
                auto_comment = parent_node.get('{}#'.format(key))
                if auto_comment:
                    auto_comments.append(auto_comment)

            message = babel_catalog.Message(
                msgid,
                None,
                auto_comments=auto_comments,
                locations=[(path, 0)])
            if msgid:
                _add_to_catalog(message, locales)

        def _babel_extract(fp, locales, path):
            try:
                all_parts = extract.extract(
                    'jinja2.ext.babel_extract',
                    fp,
                    options=options,
                    comment_tags=comment_tags)

                for parts in all_parts:
                    lineno, msgid, comments, context = parts
                    message = babel_catalog.Message(
                        msgid,
                        None,
                        auto_comments=comments,
                        locations=[(path, lineno)])
                    _add_to_catalog(message, locales)
            except tokenize.TokenError:
                self.pod.logger.error(
                    'Problem extracting body: {}'.format(path))
                raise

        # Extract from collections in /content/:
        # Strings only extracted for relevant locales, determined by locale
        # scope (pod > collection > document > document part)
        last_pod_path = None
        collection_paths = []
        for collection in self.pod.list_collections():
            collection_paths.append(collection.pod_path)

            if utils.fnmatches_paths(collection.blueprint_path, paths):
                text = 'Extracting: {}'.format(collection.blueprint_path)
                self.pod.logger.info(text)
                # Extract from blueprint.
                utils.walk(collection.tagged_fields,
                           lambda msgid, key, node, **kwargs: _handle_field(
                               collection.blueprint_path, collection.locales, msgid, key, node,
                               **kwargs))

            for doc in collection.list_docs(include_hidden=True):
                if not utils.fnmatches_paths(doc.pod_path, paths):
                    continue
                if doc.pod_path != last_pod_path:
                    self.pod.logger.info(
                        'Extracting: {} ({} locale{})'.format(
                            doc.pod_path,
                            len(doc.locales),
                            's' if len(doc.locales) != 1 else '',
                        )
                    )
                    last_pod_path = doc.pod_path

                # If doc.locale is set, this is a doc part: only extract for
                # its own locales (not those of base doc).
                if doc.locale:
                    doc_locales = [doc.locale]
                # If not is set, this is a base doc (1st or only part): extract
                # for all locales declared for this doc
                elif doc.locales:
                    doc_locales = doc.locales
                # Otherwise only include in template (--no-localized)
                else:
                    doc_locales = [None]

                doc_locales = [doc.locale]
                # Extract yaml fields: `foo@: Extract me`
                # ("tagged" = prior to stripping `@` suffix from field names)
                tagged_fields = doc.format.front_matter.raw_data
                utils.walk(tagged_fields,
                           lambda msgid, key, node, **kwargs: _handle_field(
                               doc.pod_path, doc_locales, msgid, key, node, **kwargs))

                # Extract body: {{_('Extract me')}}
                if doc.body:
                    doc_body = io.BytesIO(bytes(doc.body, 'utf-8'))
                    _babel_extract(doc_body, doc_locales, doc.pod_path)

            # Extract from CSVs for this collection's locales
            for filepath in self.pod.list_dir(collection.pod_path):
                if not utils.fnmatches_paths(filepath, paths):
                    continue
                if filepath.endswith('.csv'):
                    pod_path = os.path.join(
                        collection.pod_path, filepath.lstrip('/'))
                    self.pod.logger.info('Extracting: {}'.format(pod_path))
                    rows = self.pod.read_csv(pod_path)
                    for i, row in enumerate(rows):
                        for key, msgid in row.items():
                            _handle_field(
                                pod_path, collection.locales, msgid, key, row)

        # Extract from data directories of /content/:
        for root, dirs, _ in self.pod.walk('/content/'):
            for dir_name in dirs:
                pod_dir = os.path.join(root, dir_name)
                pod_dir = pod_dir.replace(self.pod.root, '')
                if not self._starts_with_paths(collection_paths, pod_dir):
                    for path in self.pod.list_dir(pod_dir, recursive=False):
                        if not utils.fnmatches_paths(path, paths):
                            continue

                        # Extract from non-collection csv files.
                        if path.endswith('.csv'):
                            pod_path = os.path.join(pod_dir, path.lstrip('/'))
                            self.pod.logger.info('Extracting: {}'.format(pod_path))
                            rows = self.pod.read_csv(pod_path)
                            for i, row in enumerate(rows):
                                for key, msgid in row.items():
                                    _handle_field(
                                        pod_path, self.pod.list_locales(), msgid, key, row)

                        # Extract from non-collection yaml files.
                        if path.endswith(('.yaml', '.yml')):
                            pod_path = os.path.join(pod_dir, path.lstrip('/'))
                            self.pod.logger.info('Extracting: {}'.format(pod_path))
                            utils.walk(
                                self.pod.read_yaml(pod_path),
                                lambda msgid, key, node, **kwargs: _handle_field(
                                    pod_path, self.pod.list_locales(), msgid, key, node, **kwargs)
                            )

        # Extract from data directories of /data/:
        for path in self.pod.list_dir('/data/', recursive=True):
            if not utils.fnmatches_paths(path, paths):
                continue
            if path.endswith(('.csv')):
                pod_path = os.path.join('/data/', path.lstrip('/'))
                self.pod.logger.info('Extracting: {}'.format(pod_path))
                rows = self.pod.read_csv(pod_path)
                for i, row in enumerate(rows):
                    for key, msgid in row.items():
                        _handle_field(
                            pod_path, self.pod.list_locales(), msgid, key, row)

            if path.endswith(('.yaml', '.yml')):
                pod_path = os.path.join('/data/', path.lstrip('/'))
                self.pod.logger.info('Extracting: {}'.format(pod_path))
                fields = utils.parse_yaml(
                    self.pod.read_file(pod_path), pod=self.pod)
                utils.walk(
                    fields,
                    lambda msgid, key, node, **kwargs: _handle_field(
                        pod_path, self.pod.list_locales(), msgid, key, node, **kwargs)
                )

        # Extract from root of /content/:
        for path in self.pod.list_dir('/content/', recursive=False):
            if not utils.fnmatches_paths(path, paths):
                continue
            if path.endswith(('.yaml', '.yml')):
                pod_path = os.path.join('/content/', path)
                self.pod.logger.info('Extracting: {}'.format(pod_path))
                utils.walk(
                    self.pod.get_doc(pod_path).format.front_matter.raw_data,
                    lambda msgid, key, node, **kwargs: _handle_field(
                        pod_path, self.pod.list_locales(), msgid, key, node, **kwargs)
                )

        # Extract from /views/:
        # Not discriminating by file extension, because people use all sorts
        # (htm, html, tpl, dtml, jtml, ...)
        if not audit:
            for path in self.pod.list_dir('/views/'):
                filename = os.path.basename(path)
                if not utils.fnmatches_paths(path, paths) \
                        or path.startswith(_IGNORED_PREFIXS) \
                        or filename.startswith(_IGNORED_PREFIXS):
                    continue
                pod_path = os.path.join('/views/', path)
                self.pod.logger.info('Extracting: {}'.format(pod_path))
                with self.pod.open_file(pod_path, 'rb') as f:
                    _babel_extract(f, self.pod.list_locales(), pod_path)

        # Extract from /partials/:
        if not audit:
            for path in self.pod.list_dir('/partials/'):
                filename = os.path.basename(path)
                if not utils.fnmatches_paths(path, paths) \
                        or path.startswith(_IGNORED_PREFIXS) \
                        or filename.startswith(_IGNORED_PREFIXS):
                    continue
                pod_path = os.path.join('/partials/', path)
                if path.endswith(('.yaml', '.yml')):
                    self.pod.logger.info('Extracting: {}'.format(pod_path))
                    utils.walk(
                        self.pod.get_doc(pod_path).format.front_matter.raw_data,
                        lambda msgid, key, node, **kwargs: _handle_field(
                            pod_path, self.pod.list_locales(), msgid, key, node, **kwargs)
                    )
                if path.endswith(('.html', '.htm')):
                    self.pod.logger.info('Extracting: {}'.format(pod_path))
                    with self.pod.open_file(pod_path, 'rb') as f:
                        _babel_extract(f, self.pod.list_locales(), pod_path)

        # Extract from podspec.yaml:
        if utils.fnmatches_paths('/podspec.yaml', paths):
            self.pod.logger.info('Extracting: /podspec.yaml')
            utils.walk(
                self.pod.get_podspec().get_config(),
                lambda msgid, key, node, **kwargs: _handle_field(
                    '/podspec.yaml', self.pod.list_locales(), msgid, key, node, **kwargs)
            )

        # Save it out: behavior depends on --localized and --locale flags
        # If an out path is specified, always collect strings into the one catalog.
        if localized and not out_path:
            # Save each localized catalog
            for locale, new_catalog in localized_catalogs.items():
                # Skip if `locales` defined but doesn't include this locale
                if locales and locale not in locales:
                    continue
                existing_catalog = self.get(locale)
                existing_catalog.update_using_catalog(
                    new_catalog,
                    include_obsolete=include_obsolete)
                if audit:
                    continue
                existing_catalog.save(include_header=include_header)
                missing = existing_catalog.list_untranslated()
                num_messages = len(existing_catalog)
                self.pod.logger.info(
                    'Saved: /{path} ({num_translated}/{num_messages})'.format(
                        path=existing_catalog.pod_path,
                        num_translated=num_messages - len(missing),
                        num_messages=num_messages)
                )
            return untagged_strings, localized_catalogs.items()
        else:
            # --localized omitted / --no-localized
            template_catalog = self.get_template(self.template_path)
            template_catalog.update_using_catalog(
                unlocalized_catalog,
                include_obsolete=include_obsolete)
            if not audit:
                template_catalog.save(include_header=include_header)
                text = 'Saved: {} ({} messages)'
                self.pod.logger.info(
                    text.format(template_catalog.pod_path,
                                len(template_catalog))
                )
            return untagged_strings, [template_catalog]

    def write_template(self, template_path, catalog, include_obsolete=False,
                       include_header=False):
        template_file = self.pod.open_file(template_path, mode='wb')
        catalogs.Catalog.set_header_comment(self.pod, catalog)
        pofile.write_po(
            template_file, catalog, width=80, omit_header=(not include_header),
            sort_output=True, sort_by_file=True,
            ignore_obsolete=(not include_obsolete))
        text = 'Saved: {} ({} messages)'
        self.pod.logger.info(text.format(template_path, len(catalog)))
        template_file.close()
        return catalog

    def find_mo_file(self, locale):
        identifiers = gettext._expand_lang(str(locale))
        for identifier in identifiers:
            path = os.path.join(
                '/translations', identifier, 'LC_MESSAGES',
                'messages.mo')
            try:
                return path, self.pod.open_file(path, 'rb')
            except IOError:
                pass
        return None, None

    def clear_gettext_cache(self):
        self._gettext_translations = {}

    def get_gettext_translations(self, locale):
        if locale in self._gettext_translations:
            return self._gettext_translations[locale]
        path, fp = self.find_mo_file(locale)
        if fp:
            trans = support.Translations(fp, domain='messages')
        else:
            trans = support.NullTranslations()
        self._gettext_translations[locale] = trans
        return trans

    def filter(self, out_path=None, out_dir=None,
               include_obsolete=True, localized=False,
               paths=None, include_header=None, locales=None):
        if localized and out_dir is None:
            raise UsageError('Must specify --out_dir when using --localized in '
                             'order to generate localized catalogs.')
        if not localized and out_path is None:
            raise UsageError('Must specify -o when not using --localized.')
        filtered_catalogs = []
        messages_to_locales = {}
        for locale in locales:
            locale_catalog = self.get(locale)
            missing_messages = locale_catalog.list_untranslated(paths=paths)
            num_missing = len(missing_messages)
            num_total = len(locale_catalog)
            for message in missing_messages:
                if message.id not in messages_to_locales:
                    messages_to_locales[message.id] = {
                        'message': message,
                        'locales': [locale_catalog.locale]
                    }
                else:
                    messages_to_locales[message.id]['locales'].append(
                        locale_catalog.locale)

            # Generate localized catalogs.
            if localized:
                filtered_catalog = self.get(locale, dir_path=out_dir)
                for message in missing_messages:
                    filtered_catalog[message.id] = message
                if len(filtered_catalog):
                    text = 'Saving: {} ({} missing of {})'
                    text = text.format(filtered_catalog.pod_path, num_missing,
                                       num_total)
                    self.pod.logger.info(text)
                    filtered_catalog.save(include_header=include_header)
                else:
                    text = 'Skipping: {} (0 missing of {})'
                    text = text.format(filtered_catalog.pod_path, num_total)
                    self.pod.logger.info(text)
                filtered_catalogs.append(filtered_catalog)
        if localized:
            return filtered_catalogs
        # Generate a single catalog template.
        self.pod.write_file(out_path, '')
        babel_catalog = pofile.read_po(self.pod.open_file(out_path))
        for message_id in messages_to_locales.keys():
            babel_catalog[message_id] = messages_to_locales[message_id]['message']
        self.write_template(out_path, babel_catalog,
                            include_header=include_header)
        return [babel_catalog]

    @classmethod
    def format_audit(cls, untagged_strings, extracted_catalogs):
        table = texttable.Texttable()
        table.set_deco(texttable.Texttable.HEADER)
        rows = []
        rows.append(['Location', 'String'])
        for location, string in untagged_strings:
            rows.append([location, string])
        table.add_rows(rows)
        content = table.draw()
        return content



"""Deprecated path for Grow importers."""

# TODO: Remove after deprecation period.

from grow.common import deprecated
from grow.translations import importers as new_ref

# pylint: disable=invalid-name
Error = deprecated.MovedHelper(
    new_ref.Error, 'grow.pods.importers.Error')
Importer = deprecated.MovedHelper(
    new_ref.Importer, 'grow.pods.importers.Importer')



"""Subcommand for displaying untranslated strings in the pod."""

import os
import click
from grow.commands import shared
from grow.common import rc_config
from grow.pods import pods
from grow import storage
from grow.rendering import renderer


CFG = rc_config.RC_CONFIG.prefixed('grow.inspect.untranslated')


@click.command(name='untranslated')
@shared.pod_path_argument
@shared.locale_option(
    help_text='Which locale(s) to analyze when searching for untranslated strings.')
@shared.localized_option(CFG)
@shared.threaded_option(CFG)
def inspect_untranslated(pod_path, locale, localized, threaded):
    """Displays statistics about the pod."""
    root = os.path.abspath(os.path.join(os.getcwd(), pod_path))
    pod = pods.Pod(root, storage=storage.FileStorage)
    with pod.profile.timer('grow_inspect_untranslated'):
        # Find all of the untagged strings.
        catalogs = pod.get_catalogs(template_path=None)
        untagged_strings, _ = catalogs.extract(
            localized=localized, locales=locale, audit=True, out_path=None)

        # Find all the untranslated strings in use.
        pod.enable(pod.FEATURE_TRANSLATION_STATS)

        pod.router.use_simple()
        pod.router.add_all()
        if locale:
            pod.router.filter('whitelist', locales=list(locale))
        content_generator = renderer.Renderer.rendered_docs(
            pod, pod.router.routes, use_threading=threaded)

        # Render each document to get the untranslated strings.
        for _ in content_generator:
            pass

        pod.translation_stats.add_untagged(untagged_strings)
        pod.translation_stats.pretty_print()
    return pod



"""Router for grow documents."""

import fnmatch
import os
import re
from protorpc import messages
from grow.performance import docs_loader
from grow.rendering import render_controller
from grow.routing import path_filter as grow_path_filter
from grow.routing import routes as grow_routes


class FilterConfig(messages.Message):
    """Configuration for routing filters in podspec."""
    type = messages.StringField(1)
    collections = messages.StringField(2, repeated=True)
    paths = messages.StringField(3, repeated=True)
    locales = messages.StringField(4, repeated=True)
    kinds = messages.StringField(5, repeated=True)


class Error(Exception):
    """Base router error."""

    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class MissingStaticConfigError(Error):
    """Error for missing a configuration for a static file."""
    pass


class Router(object):
    """Router for pods."""

    def __init__(self, pod, routes=None):
        self.pod = pod
        self._routes = routes or grow_routes.Routes()

    def _preload_and_expand(self, docs, expand=True):
        # Force preload the docs.
        docs_loader.DocsLoader.load(self.pod, docs)
        docs_loader.DocsLoader.fix_default_locale(self.pod, docs)
        if expand:
            # Will need all of the docs, so expand them out and preload.
            docs = docs_loader.DocsLoader.expand_locales(self.pod, docs)
            docs_loader.DocsLoader.load(self.pod, docs)
        return docs

    @property
    def routes(self):
        """Routes reflective of the docs."""
        return self._routes

    def _add_to_routes(self, path, route_info, concrete=True, options=None, fingerprinted=None):
        """Add to the routes and the route cache."""
        self.routes.add(path, route_info, options=options)

        # Don't want to keep track of the concrete fingerprinted routes.
        if not concrete or not fingerprinted:
            self.pod.podcache.routes_cache.add(
                path, route_info, options=options, concrete=concrete,
                env=self.pod.env.name)

    def add_all(self, concrete=True, use_cache=True):
        """Add all documents and static content."""
        if use_cache:
            unchanged_pod_paths = self.from_cache(concrete=concrete)
        else:
            unchanged_pod_paths = []

        self.add_all_docs(
            concrete=concrete, unchanged_pod_paths=unchanged_pod_paths)
        self.add_all_static(
            concrete=concrete, unchanged_pod_paths=unchanged_pod_paths)
        self.add_all_other(concrete=concrete)
        self.add_all_hook(concrete=concrete)

    def add_all_docs(self, concrete=True, unchanged_pod_paths=None):
        """Add all pod docs to the router."""
        with self.pod.profile.timer('Router.add_all_docs'):
            unchanged_pod_paths = unchanged_pod_paths or set()
            docs = []
            for collection in self.pod.list_collections():
                doc_basenames = set()
                for doc in collection.list_docs_unread():
                    # Skip when the doc is in the unchanged pod paths set.
                    if doc.pod_path in unchanged_pod_paths:
                        # Even though it is being skipped, mark that it has been
                        # loaded to prevent issues with paramaterized serving paths.
                        doc_basenames.add(doc.collection_sub_path_clean)
                        continue
                    # Skip duplicate documents when using non-concrete routing.
                    if not concrete and doc.collection_sub_path_clean in doc_basenames:
                        continue
                    is_default_locale = doc._locale_kwarg == self.pod.podspec.default_locale
                    # Ignore localized names in the files since they will be
                    # picked up when the locales are expanded.
                    if doc.root_pod_path == doc.pod_path or is_default_locale:
                        docs.append(doc)
                        doc_basenames.add(doc.collection_sub_path_clean)
                    else:
                        # If this document does not exist with the default
                        # locale it still needs to be added.
                        locale_doc = self.pod.get_doc(
                            doc.pod_path, doc.default_locale)
                        if not locale_doc.exists:
                            docs.append(doc)
                            doc_basenames.add(doc.collection_sub_path_clean)
            docs = self._preload_and_expand(docs, expand=concrete)
            self.add_docs(docs, concrete=concrete)

    def add_all_hook(self, concrete=True):
        """Trigger the hook for adding all to the routes."""
        with self.pod.profile.timer('Router.add_all_hook'):
            self.pod.extensions_controller.trigger(
                'router_add', self, concrete=concrete)

    def add_all_other(self, concrete=True):
        """Add all pod docs to the router."""
        with self.pod.profile.timer('Router.add_all_other'):
            podspec = self.pod.podspec.get_config()
            if 'error_routes' in podspec:
                for key, error_route in podspec['error_routes'].items():
                    if key == 'default':
                        key = 404

                    path = '/{}.html'.format(key)
                    route_info = RouteInfo('error', meta={
                        'key': key,
                        'view': error_route,
                    })
                    self._add_to_routes(path, route_info, concrete=concrete)

            if 'sitemap' in podspec:
                sitemap = podspec['sitemap']
                default_sitemap_path = self.pod.podspec.root + '/sitemap.xml'
                default_sitemap_path = default_sitemap_path.replace('//', '/')
                sitemap_path = self.pod.path_format.format_pod(
                    sitemap.get('path', default_sitemap_path))
                route_info = RouteInfo('sitemap', meta={
                    'collections': sitemap.get('collections'),
                    'locales': sitemap.get('locales'),
                    'template': sitemap.get('template'),
                    'filters': sitemap.get('filters'),
                    'path': sitemap_path,
                })
                self._add_to_routes(
                    sitemap_path, route_info, concrete=concrete)

    def add_all_static(self, concrete=True, unchanged_pod_paths=None):
        """Add all pod docs to the router."""
        with self.pod.profile.timer('Router.add_all_static'):
            unchanged_pod_paths = unchanged_pod_paths or set()
            skipped_paths = []
            for config in self.pod.static_configs:
                if config.get('dev') and not self.pod.env.dev:
                    continue

                fingerprinted = config.get('fingerprinted', False)
                localization = config.get('localization')
                static_filter = config.get('filter', {})
                if static_filter:
                    path_filter = grow_path_filter.PathFilter(
                        static_filter.get('ignore_paths'),
                        static_filter.get('include_paths'))
                else:
                    path_filter = self.pod.path_filter

                static_dirs = config.get('static_dirs')
                if not static_dirs:
                    static_dirs = [config.get('static_dir')]

                if localization:
                    localized_static_dirs = localization.get('static_dirs')
                    if not localized_static_dirs:
                        localized_static_dirs = [
                            localization.get('static_dir')]

                if concrete:
                    # Enumerate static files.
                    for static_dir in static_dirs:
                        for root, dirs, files in self.pod.walk(static_dir):
                            for directory in dirs:
                                if directory.startswith('.'):
                                    dirs.remove(directory)
                            pod_dir = root.replace(self.pod.root, '')
                            for file_name in files:
                                pod_path = os.path.join(pod_dir, file_name)
                                # Skip when the doc is in the unchanged pod paths set.
                                if pod_path in unchanged_pod_paths:
                                    continue
                                static_doc = self.pod.get_static(
                                    pod_path, locale=None)
                                self.add_static_doc(
                                    static_doc, concrete=concrete)
                    if localization:
                        # TODO handle the localized static files?
                        pass
                else:
                    serve_at = self.pod.path_format.format_pod(
                        config['serve_at'], parameterize=True)
                    route_info = RouteInfo('static', meta={
                        'path_format': serve_at,
                        'source_formats': static_dirs,
                        'localized': False,
                        'localization': localization,
                        'fingerprinted': fingerprinted,
                        'static_filter': static_filter,
                        'path_filter': path_filter,
                    })
                    self._add_to_routes(
                        serve_at + '*', route_info, concrete=concrete,
                        fingerprinted=fingerprinted)

                    if localization:
                        localized_serve_at = self.pod.path_format.format_pod(
                            localization.get('serve_at'), parameterize=True)
                        route_info = RouteInfo('static', meta={
                            'path_format': localized_serve_at,
                            'source_formats': localized_static_dirs,
                            'localized': True,
                            'localization': localization,
                            'fingerprinted': fingerprinted,
                            'static_filter': static_filter,
                            'path_filter': path_filter,
                        })
                        self._add_to_routes(
                            localized_serve_at + '*', route_info, concrete=concrete,
                            fingerprinted=fingerprinted)
            if skipped_paths:
                self.pod.logger.info(
                    'Ignored {} static files.'.format(len(skipped_paths)))

    def add_doc(self, doc, concrete=True):
        """Add doc to the router."""
        if not doc.has_serving_path():
            return
        route_info = RouteInfo(
            'doc', pod_path=doc.pod_path,
            hashed=self.pod.hash_file(doc.pod_path),
            meta={
                'pod_path': doc.pod_path,
                'locale': str(doc.locale),
                'collection_path': doc.collection.pod_path,
            })
        self._add_to_routes(
            doc.get_serving_path(), route_info,
            concrete=concrete, options=doc.path_params)

    def add_docs(self, docs, concrete=True):
        """Add docs to the router."""
        with self.pod.profile.timer('Router.add_docs'):
            skipped_paths = []
            for doc in docs:
                if not doc.has_serving_path():
                    continue
                if not self.pod.path_filter.is_valid(doc.get_serving_path()):
                    skipped_paths.append(doc.get_serving_path())
                    continue
                if concrete:
                    # Concrete iterates all possible documents.
                    route_info = RouteInfo(
                        'doc', pod_path=doc.pod_path,
                        hashed=self.pod.hash_file(doc.pod_path),
                        meta={
                            'pod_path': doc.pod_path,
                            'locale': str(doc.locale),
                            'collection_path': doc.collection.pod_path,
                        })
                    self._add_to_routes(
                        doc.get_serving_path(), route_info,
                        options=doc.path_params, concrete=concrete)
                else:
                    # Use the raw paths to parameterize the routing.
                    base_path = doc.get_serving_path_base()
                    only_localized = doc.path_format == doc.path_format_localized
                    # If the path is already localized only add the
                    # parameterized version of the path.
                    if base_path and not only_localized:
                        route_info = RouteInfo(
                            'doc', pod_path=doc.pod_path,
                            hashed=self.pod.hash_file(doc.pod_path),
                            meta={
                                'pod_path': doc.pod_path,
                                'locale': str(doc.locale),
                                'collection_path': doc.collection.pod_path,
                            })
                        self._add_to_routes(
                            base_path, route_info,
                            concrete=concrete, options=doc.path_params)
                    localized_path = doc.get_serving_path_localized()
                    if not localized_path or ':locale' not in localized_path:
                        localized_paths = doc.get_serving_paths_localized()
                        for locale, path in localized_paths.items():
                            route_info = RouteInfo(
                                'doc', pod_path=doc.pod_path,
                                hashed=self.pod.hash_file(doc.pod_path),
                                meta={
                                    'pod_path': doc.pod_path,
                                    'locale': str(locale),
                                    'collection_path': doc.collection.pod_path,
                                })
                            self._add_to_routes(
                                path, route_info,
                                concrete=concrete, options=doc.path_params)
                    else:
                        route_info = RouteInfo(
                            'doc', pod_path=doc.pod_path,
                            hashed=self.pod.hash_file(doc.pod_path),
                            meta={
                                'pod_path': doc.pod_path,
                                'collection_path': doc.collection.pod_path,
                            })
                        self._add_to_routes(
                            localized_path, route_info,
                            concrete=concrete, options=doc.path_params_localized)
            if skipped_paths:
                self.pod.logger.info(
                    'Ignored {} documents.'.format(len(skipped_paths)))

    def add_pod_paths(self, pod_paths, concrete=True):
        """Add pod paths to the router."""
        with self.pod.profile.timer('Router.add_pod_paths'):
            # Find all of the doc pod_path for matching.
            all_doc_pod_paths = set()
            for collection in self.pod.list_collections():
                for doc in collection.list_docs_unread():
                    all_doc_pod_paths.add(doc.pod_path)

            doc_pod_paths = set()
            for pod_path in pod_paths:
                # Add docs from the depdency graph.
                dependents = self.pod.podcache.dependency_graph.match_dependents(
                    pod_path)
                for dep_path in dependents:
                    doc_pod_paths.add(dep_path)

                # Add docs based just on the doc pod paths.
                for doc_pod_path in all_doc_pod_paths:
                    if fnmatch.fnmatch(doc_pod_path, pod_path):
                        doc_pod_paths.add(doc_pod_path)

            docs = []
            for pod_path in doc_pod_paths:
                docs.append(self.pod.get_doc(pod_path))
            docs = self._preload_and_expand(docs, expand=concrete)
            self.add_docs(docs, concrete=concrete)

    def add_static_doc(self, static_doc, concrete=True):
        """Add static doc to the router."""
        if not static_doc.path_filter.is_valid(static_doc.serving_path):
            return
        route_info = RouteInfo(
            'static', pod_path=static_doc.pod_path,
            hashed=self.pod.hash_file(static_doc.pod_path),
            meta={
                'pod_path': static_doc.pod_path,
                'locale': None,
                'localized': False,
                'localization': static_doc.config.get('localization'),
                'fingerprinted': static_doc.fingerprinted,
                'static_filter': static_doc.filter,
                'path_filter': static_doc.path_filter,
            })
        self._add_to_routes(
            static_doc.serving_path, route_info, concrete=concrete,
            fingerprinted=static_doc.fingerprinted)

    def filter(self, filter_type, collection_paths=None, paths=None, locales=None, kinds=None):
        """Filter the routes based on the filter type and criteria."""
        # Convert paths to be regex.
        regex_paths = []
        for path in paths or []:
            regex_paths.append(re.compile(path))

        # Ability to specify a none locale using commanf flag.
        locales = locales or []
        if 'None' in locales:
            locales.append(None)

        if filter_type == 'whitelist':
            def _filter_whitelist(serving_path, route_info):
                # Check for whitelisted collection path.
                if collection_paths and 'collection_path' in route_info.meta:
                    if route_info.meta['collection_path'] in collection_paths:
                        return True

                # Check for whitelisted serving path.
                for regex_path in regex_paths:
                    if regex_path.match(serving_path):
                        return True

                # Check for whitelisted locale.
                if 'locale' in route_info.meta:
                    if route_info.meta['locale'] in locales:
                        return True

                # Check for whitelisted kind of routes.
                if kinds and route_info.kind in kinds:
                    return True

                # Not whitelist.
                return False

            count = self.routes.filter(_filter_whitelist)
            if count > 0:
                self.pod.logger.info('Whitelist filtered out {} routes.'.format(count))
        else:
            def _filter_blacklist(serving_path, route_info):
                # Check for blacklisted collection path.
                if collection_paths and 'collection_path' in route_info.meta:
                    if route_info.meta['collection_path'] in collection_paths:
                        return False

                # Check for blacklisted serving path.
                for regex_path in regex_paths:
                    if regex_path.match(serving_path):
                        return False

                # Check for blacklisted locale.
                if 'locale' in route_info.meta:
                    if route_info.meta['locale'] in locales:
                        return False

                # Check for blacklisted kinds of routes.
                if kinds and route_info.kind in kinds:
                    return False

                # Not blacklisted.
                return True

            count = self.routes.filter(_filter_blacklist)
            if count > 0:
                self.pod.logger.info('Blacklist filtered out {} routes.'.format(count))

    def from_cache(self, concrete=True):
        """Import routes from routes cache."""
        fingerprinted_dirs = set()
        for config in self.pod.static_configs:
            if 'fingerprinted' in config and config['fingerprinted']:
                fingerprinted_dirs.add(config['static_dir'])
        fingerprinted_dirs = tuple(fingerprinted_dirs)

        routes_data = self.pod.podcache.routes_cache.raw(
            concrete=concrete, env=self.pod.env.name)
        unchanged_pod_paths = set()
        removed_paths = []
        for key, item in routes_data.items():
            route_info = item['value']

            # For now ignore anything that doesn't have a hash.
            if not route_info.hashed:
                continue

            # If the serving path is no longer in the static configs, ignore.
            if route_info.kind == 'static':
                try:
                    self.get_static_config_for_serving_path(key)
                except MissingStaticConfigError:
                    continue

            # Ignore deleted files.
            if not self.pod.file_exists(route_info.pod_path):
                removed_paths.append(key)
                continue

            # If the hash has changed then skip.
            if route_info.hashed != self.pod.hash_file(route_info.pod_path):
                continue

            # Ignore the fingerprinted files.
            if 'fingerprinted' in route_info.meta:
                if route_info.meta['fingerprinted']:
                    continue
                else:
                    # Check if the configuration has changed and the path should
                    # be fingerprinted.
                    if route_info.pod_path.startswith(fingerprinted_dirs):
                        continue

            unchanged_pod_paths.add(route_info.pod_path)
            self.routes.add(key, route_info, options=item['options'])

        for key in removed_paths:
            self.pod.podcache.routes_cache.remove(key, concrete=concrete)

        return unchanged_pod_paths

    def from_data(self, routes_data):
        """Import routes from data."""
        for key, item in routes_data.items():
            self.routes.add(
                key,
                RouteInfo.from_data(**item['value']),
                options=item['options'])

    def get_render_controller(self, path, route_info, params=None):
        """Find the correct render controller for the given route info."""
        return render_controller.RenderController.from_route_info(
            self.pod, path, route_info, params=params)

    def get_static_config_for_pod_path(self, pod_path):
        """Return the static configuration for a pod path."""
        text = '{} is not found in any static file configuration in the podspec.'
        if pod_path is None:
            raise MissingStaticConfigError(text.format(pod_path))

        for config in self.pod.static_configs:
            if config.get('dev') and not self.pod.env.dev:
                continue
            static_dirs = config.get('static_dirs')
            if not static_dirs:
                static_dirs = [config.get('static_dir')]
            if isinstance(static_dirs, str):
                static_dirs = [static_dirs]
            for static_dir in static_dirs:
                if pod_path.startswith(static_dir):
                    return config
            intl = config.get('localization', {})
            if intl:
                static_dirs = intl.get('static_dirs')
                if not static_dirs:
                    static_dirs = [intl.get('static_dir')]
                if isinstance(static_dirs, str):
                    static_dirs = [static_dirs]
                for static_dir in static_dirs:
                    if pod_path.startswith(static_dir):
                        return config

        raise MissingStaticConfigError(text.format(pod_path))

    def get_static_config_for_serving_path(self, serving_path):
        """Return the static configuration for a pod path."""
        text = '{} is not found in any static file configuration in the podspec.'
        if serving_path is None:
            raise MissingStaticConfigError(text.format(serving_path))

        for config in self.pod.static_configs:
            if config.get('dev') and not self.pod.env.dev:
                continue
            serve_at = config.get('serve_at')
            if serving_path.startswith(serve_at):
                return config
            intl = config.get('localization', {})
            if intl:
                serve_at = intl.get('serve_at')
                if serving_path.startswith(serve_at):
                    return config

        raise MissingStaticConfigError(text.format(serving_path))

    def reconcile_documents(self, remove_docs=None, add_docs=None):
        """Remove old docs and add new docs to the routes."""
        for doc in remove_docs if remove_docs else []:
            self.routes.remove(doc.get_serving_path())
        for doc in add_docs if add_docs else []:
            self.add_doc(doc)

    def shard(self, shard_count, current_shard, attr='kind'):
        """Removes paths from the routes based on sharding rules."""
        self.routes.shard(shard_count, current_shard, attr=attr)

    def use_simple(self):
        """Switches the routes to be a simple routes object."""
        previous_routes = self._routes
        self._routes = grow_routes.RoutesSimple()
        if previous_routes is not None:
            for path, value, _ in previous_routes.nodes:
                self._routes.add(path, value)


# pylint: disable=too-few-public-methods
class RouteInfo(object):
    """Organize information stored in the routes."""

    def __eq__(self, other):
        return self.kind == other.kind and self.meta == other.meta

    def __init__(self, kind, pod_path=None, hashed=None, meta=None):
        self.kind = kind
        self.pod_path = pod_path
        self.hashed = hashed
        self.meta = meta or {}

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '<RouteInfo kind={} meta={}>'.format(self.kind, self.meta)

    @classmethod
    def from_data(cls, kind, pod_path, hashed, meta):
        """Create the route info from data."""
        # Need to reconstruct the path filter object.
        if 'path_filter' in meta:
            meta['path_filter'] = grow_path_filter.PathFilter(
                ignored=meta['path_filter'].get('ignored'),
                included=meta['path_filter'].get('included'))
        if 'static_filter' in meta:
            meta['static_filter'] = grow_path_filter.PathFilter(
                ignored=meta['static_filter'].get('ignored'),
                included=meta['static_filter'].get('included'))
        return cls(kind, pod_path=pod_path, hashed=hashed, meta=meta)

    def export(self):
        """Export route information in a serializable format."""
        export_meta = {}
        for key in self.meta:
            meta_value = self.meta[key]
            if hasattr(meta_value, 'export'):
                meta_value = meta_value.export()
            export_meta[key] = meta_value
        return {
            'kind': self.kind,
            'pod_path': self.pod_path,
            'hashed': self.hashed,
            'meta': export_meta,
        }



"""Extension controller for working with extensions in the pod."""

from grow.extensions import core
from grow.extensions import extension_importer
from grow.extensions import hooks
from grow.extensions import hook_controller

class ExtensionController(object):
    """Controller for working with pod extensions."""

    def __init__(self, pod):
        self.pod = pod
        self._hooks = {}
        self._extensions = {}

        for hook in hooks.HOOKS:
            self._hooks[hook.KEY] = hook_controller.HookController(
                self.pod, hook.KEY, hook(None))

    def extension_config(self, extension_path):
        if extension_path not in self._extensions:
            return {}
        return  self._extensions[extension_path].config

    def register_builtins(self):
        """Add new built-in extensions."""
        new_extensions = []
        for extension_cls in core.EXTENSIONS:
            ext = extension_cls(self.pod, {})
            new_extensions.append(ext)

        # Register the hooks with the hook controllers.
        for _, hook in self._hooks.items():
            hook.register_extensions(new_extensions)

    def register_extensions(self, extension_configs):
        """Add new extensions to the controller."""
        new_extensions = []
        for config_item in extension_configs:
            if isinstance(config_item, str):
                extension_path = config_item
                config = {}
            else:
                extension_path = list(config_item.keys())[0]
                config = config_item[extension_path]
            cls = extension_importer.ExtensionImporter.find_extension(
                extension_path, self.pod.root)
            ext = cls(self.pod, config)
            new_extensions.append(ext)
            self._extensions[extension_path] = ext

        # Register the hooks with the hook controllers.
        for _, hook in self._hooks.items():
            hook.register_extensions(new_extensions)

    def trigger(self, hook_key, *args, **kwargs):
        """Trigger a hook."""
        return self._hooks[hook_key].trigger(*args, **kwargs)

    def update_extension_configs(self, extension_configs):
        """Update existing extensions with new configs."""
        for config_item in extension_configs:
            if isinstance(config_item, str):
                extension_path = config_item
                config = {}
            else:
                extension_path = list(config_item.keys())[0]
                config = config_item[extension_path]
            ext = self._extensions[extension_path]
            ext.config = config



"""Grow prerequisite installers."""

from grow.sdk.installers import bower_installer
from grow.sdk.installers import extensions_installer
from grow.sdk.installers import gerrit_installer
from grow.sdk.installers import gulp_installer
from grow.sdk.installers import npm_installer
from grow.sdk.installers import nvm_installer

# The order here determines the order the installers are run.
BUILT_IN_INSTALLERS = [
    extensions_installer.ExtensionsInstaller,
    gerrit_installer.GerritInstaller,
    nvm_installer.NvmInstaller,  # Before any node based installers.
    npm_installer.NpmInstaller,
    bower_installer.BowerInstaller,  # After the npm install.
    gulp_installer.GulpInstaller,  # After the npm install.
]



import os
import sys

# Allows "import grow" and "from grow import <name>".
sys.path.extend([os.path.join(os.path.dirname(__file__), '..')])

# Set up path for make test-gae.
if 'NOSEGAE' in os.environ:
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'lib')))

# Allow `import grow; grow.Pod`.
from .pods.pods import Pod

# Allow `import grow; grow.Preprocessor` for custom preprocessors.
from .preprocessors.base import BasePreprocessor as Preprocessor

# Allow `import grow; grow.Translator` for custom translators.
from .translators.base import Translator



"""Subcommand for inspecting pods."""

import click
from grow.commands.subcommands import inspect_routes
from grow.commands.subcommands import inspect_stats
from grow.commands.subcommands import inspect_untranslated


@click.group()
def inspect():
    """Inspect pod files and stats."""
    pass


# Add the sub commands.
inspect.add_command(inspect_routes.inspect_routes)
inspect.add_command(inspect_stats.inspect_stats)
inspect.add_command(inspect_untranslated.inspect_untranslated)



"""Base class for writing hooks."""


class BaseHook(object):
    """Base class for all hooks."""

    KEY = 'hook'
    NAME = 'Hook'

    def __init__(self, extension):
        self.extension = extension

    @property
    def pod(self):
        """Reference to the pod."""
        return self.extension.pod

    # pylint:disable=no-self-use, unused-argument
    def should_trigger(self, previous_result, *_args, **_kwargs):
        """Determine if the hook should trigger."""
        return True

    def trigger(self, previous_result, *_args, **_kwargs):
        """Trigger the hook."""
        raise NotImplementedError()



"""
    Google Drive (Sheets and Docs) preprocessors allow you to store content in
    Google Drive and bring it into Grow. Grow will authenticate to the Google
    Drive API using OAuth2 and then download content as specified in
    `podspec.yaml`.

    Grow supports various ways to transform the content, e.g. Sheets can be
    downloaded and converted to yaml, and Docs can be downloaded and converted
    to markdown.
"""

import io
import csv
import json
import logging
import os
import httplib2
import slugify
from googleapiclient import discovery
from googleapiclient import errors
from protorpc import messages
from grow.common import oauth
from grow.common import untag
from grow.common import utils
from grow.documents import document_format
from grow.documents import document_front_matter as doc_front_matter
from grow.preprocessors import base


OAUTH_SCOPE = 'https://www.googleapis.com/auth/drive'
STORAGE_KEY = 'Grow SDK'
META_KEY = '$meta'
DRAFT_KEY = '$draft'
IGNORE_INITIAL = ('_', '#', '*')


# Silence extra logging from googleapiclient.
discovery.logger.setLevel(logging.WARNING)


class BaseGooglePreprocessor(base.BasePreprocessor):

    @staticmethod
    def create_service(api='drive', version='v2'):
        credentials = oauth.get_or_create_credentials(
            scope=OAUTH_SCOPE, storage_key='Grow SDK')
        http = httplib2.Http()
        http = credentials.authorize(http)
        return discovery.build(api, version, http=http)

    def run(self, build=True):
        try:
            self.execute(self.config)
        except errors.HttpError as e:
            self.logger.error(str(e))

    def execute(self, config):
        raise NotImplementedError


class GoogleDocsPreprocessor(BaseGooglePreprocessor):
    KIND = 'google_docs'
    _edit_url_format = 'https://docs.google.com/document/d/{id}/edit'

    class Config(messages.Message):
        path = messages.StringField(1)
        id = messages.StringField(2)
        convert = messages.BooleanField(3)
        folder = messages.StringField(4)
        collection = messages.StringField(5)

    @classmethod
    def download(cls, path, doc_id, logger=None, raise_errors=False):
        logger = logger or logging
        service = BaseGooglePreprocessor.create_service()
        # pylint: disable=no-member
        resp = service.files().get(fileId=doc_id).execute()
        if 'exportLinks' not in resp:
            text = 'Unable to export Google Doc: {}'
            logger.error(text.format(path))
            logger.error('Received: {}'.format(resp))
            return
        for mimetype, url in resp['exportLinks'].items():
            if not mimetype.endswith('html'):
                continue
            resp, content = service._http.request(url)
            if resp.status != 200:
                text = 'Error {} downloading Google Doc: {}'
                text = text.format(resp.status, path)
                if raise_errors:
                    raise base.PreprocessorError(text)
                logger.error(text)
            return content
        if raise_errors:
            text = 'No file to export from Google Docs: {}'.format(path)
            raise base.PreprocessorError(text)

    @classmethod
    def format_content(cls, path, content, convert=True, existing_data=None):
        ext = os.path.splitext(path)[1]
        convert_to_markdown = ext == '.md' and convert is not False
        content = utils.clean_html(
            content, convert_to_markdown=convert_to_markdown)
        # Preserve any existing frontmatter, return new content.
        if existing_data:
            if doc_front_matter.BOUNDARY_REGEX.search(existing_data):
                front_matter, old_content = doc_front_matter.DocumentFrontMatter.split_front_matter(
                    existing_data)
                return document_format.DocumentFormat.format_doc(front_matter, content)
        return content

    def _execute_doc(self, path, doc_id, convert):
        content = GoogleDocsPreprocessor.download(
            path, doc_id=doc_id, logger=self.pod.logger)
        existing_data = None
        if self.pod.file_exists(path):
            existing_data = self.pod.read_file(path)
        content = GoogleDocsPreprocessor.format_content(
            path, content, convert=convert, existing_data=existing_data)
        self.pod.write_file(path, content)
        self.logger.info('Downloaded Google Doc -> {}'.format(path))

    def execute(self, config):
        convert = config.convert is not False
        # Binds a Google Drive folder to a collection.
        if config.folder:
            service = BaseGooglePreprocessor.create_service()
            query = "'{}' in parents".format(config.folder)
            # pylint: disable=no-member
            resp = service.files().list(q=query).execute()
            docs_to_add = []
            existing_docs = self.pod.list_dir(config.collection)
            for item in resp['items']:
                doc_id = item['id']
                title = item['title']
                if item['mimeType'] != 'application/vnd.google-apps.document':
                    continue
                if title.startswith(IGNORE_INITIAL):
                    self.pod.logger.info('Skipping -> {}'.format(title))
                    continue
                if self.pod.is_enabled(self.pod.FEATURE_OLD_SLUGIFY):
                    basename = '{}.md'.format(utils.slugify(title))
                else:
                    basename = '{}.md'.format(slugify.slugify(title))
                docs_to_add.append(basename)
                path = os.path.join(config.collection, basename)
                self._execute_doc(path, doc_id, convert)
            # Clean up files that are no longer in Google Drive.
            for path in existing_docs:
                doc_path = path.lstrip(os.path.sep)
                if doc_path.startswith(IGNORE_INITIAL):
                    continue
                if doc_path not in docs_to_add:
                    path_to_delete = os.path.join(config.collection, doc_path)
                    text = 'Deleting -> {}'.format(path_to_delete)
                    self.pod.logger.info(text)
                    self.pod.delete_file(path_to_delete)
            return
        # Downloads a single document.
        doc_id = config.id
        path = config.path
        self._execute_doc(config.path, doc_id, convert)

    def get_edit_url(self, doc=None):
        """Returns the URL to edit in Google Docs."""
        return GoogleDocsPreprocessor._edit_url_format.format(id=self.config.id)


class GoogleSheetsPreprocessor(BaseGooglePreprocessor):
    KIND = 'google_sheets'
    GRID_TYPES = ['grid']
    MAP_TYPES = ['map', 'strings']
    _sheet_edit_url_format = 'https://docs.google.com/spreadsheets/d/{id}/edit'
    _edit_url_format = 'https://docs.google.com/spreadsheets/d/{id}/edit#gid={gid}'

    class Config(messages.Message):
        path = messages.StringField(1)
        id = messages.StringField(2)
        gid = messages.IntegerField(3)
        output_style = messages.StringField(4, default='compressed')
        format = messages.StringField(5, default='list')
        preserve = messages.StringField(6, default='builtins')
        gids = messages.IntegerField(7, repeated=True)
        collection = messages.StringField(8)
        output_format = messages.StringField(9, default='yaml')
        generate_ids = messages.BooleanField(10, default=False)
        header_row_count = messages.IntegerField(11, default=1)
        header_row_index = messages.IntegerField(12, default=1)
        include_properties = messages.StringField(13, repeated=True)
        color_as_draft = messages.BooleanField(14, default=True)
        keep_empty_values = messages.BooleanField(15, default=False)

    @staticmethod
    def _convert_rows_to_mapping(reader):
        results = {}

        def _update_node(root, part):
            if isinstance(root, dict) and part not in root:
                root[part] = {}
        for row in reader:
            key = row[0]
            value = row[1]
            if key.startswith(IGNORE_INITIAL):
                continue
            if '.' in key:
                parts = key.split('.')
                parent = results
                for i, part in enumerate(parts):
                    _update_node(parent, part)
                    if i + 1 < len(parts):
                        parent = parent[part]
                if isinstance(parent, dict):
                    parent[part] = value
            else:
                results[key] = value
        return results

    @staticmethod
    def column_to_letter(column_num):
        div = column_num
        string = ""
        temp = 0
        while div > 0:
            module = (div - 1) % 26
            string = chr(65 + module) + string
            div = int((div - module) / 26)
        return string

    @staticmethod
    def format_as_map(fp):
        reader = csv.reader(fp)
        results = GoogleSheetsPreprocessor._convert_rows_to_mapping(reader)
        return results

    @classmethod
    def download(cls, spreadsheet_id, gids=None, format_as='list', logger=None,
                 generate_ids=False, header_row_count=1, header_row_index=1,
                 keep_empty_values=False):
        logger = logger or logging
        # Show metadata about the file to help the user better understand what
        # they are downloading. Also include a link in the output to permit the
        # user to quickly open the file.
        drive_service = BaseGooglePreprocessor.create_service('drive', 'v3')
        # pylint: disable=no-member
        resp = drive_service.files().get(
            fileId=spreadsheet_id,
            fields='name,modifiedTime,lastModifyingUser,webViewLink').execute()
        title = resp['name']
        if 'lastModifyingUser' in resp:
            # Sometimes the email address isn't included.
            name = resp['lastModifyingUser']['displayName']
            modified_by = resp['lastModifyingUser'].get('emailAddress', name)
            logger.info('Downloading "{}" modified {} by {} from {}'.format(
                title,
                resp['modifiedTime'],
                modified_by,
                resp['webViewLink']))
        else:
            logger.info('Downloading "{}" modified {} from {}'.format(
                title,
                resp['modifiedTime'],
                resp['webViewLink']))

        service = BaseGooglePreprocessor.create_service('sheets', 'v4')
        format_as_grid = format_as in cls.GRID_TYPES
        format_as_map = format_as in cls.MAP_TYPES
        # pylint: disable=no-member
        spreadsheet = service.spreadsheets().get(
            spreadsheetId=spreadsheet_id).execute()

        gid_to_sheet = {}
        for sheet in spreadsheet['sheets']:
            gid_to_sheet[sheet['properties']['sheetId']] = sheet['properties']

        if not gids:
            gids = list(gid_to_sheet.keys())
        if gids and len(gids) > 1:
            url = GoogleSheetsPreprocessor._sheet_edit_url_format.format(
                id=spreadsheet_id)
            logger.info('Downloading {} tabs -> {}'.format(len(gids), url))

        gids_to_process = []
        for gid in gids:
            title = gid_to_sheet[gid]['title']
            if title.startswith(IGNORE_INITIAL):
                logger.info('Skipping tab -> {}'.format(title))
                continue
            gids_to_process.append(gid)

        gid_to_data = {}
        generated_key_index = 0

        range_names = []

        for gid in gids_to_process:
            if format_as_map:
                max_column = 'B'
            else:
                max_column = GoogleSheetsPreprocessor.column_to_letter(
                    gid_to_sheet[gid]['gridProperties']['columnCount'])
            range_name = "'{}'!A:{}".format(
                gid_to_sheet[gid]['title'], max_column)
            range_names.append(range_name)

        # pylint: disable=no-member
        batch_resp = service.spreadsheets().values().batchGet(
            spreadsheetId=spreadsheet_id, ranges=range_names).execute()

        for i, gid in enumerate(gids_to_process):
            resp = batch_resp['valueRanges'][i]
            if format_as_map or format_as_grid:
                gid_to_data[gid] = {}
            else:
                gid_to_data[gid] = []

            if not 'values' in resp:
                logger.info(
                    'No values found in sheet -> {}'.format(gid_to_sheet[gid]['title']))
            else:
                headers = None
                header_rows = []
                for row in resp['values']:
                    if len(header_rows) < header_row_count:
                        header_rows.append(row)
                        # Only one of the header rows are the actual headers.
                        if len(header_rows) == header_row_index:
                            if format_as_grid:
                                # Ignore first column as a header.
                                headers = row[1:]
                            else:
                                headers = row
                        continue

                    if format_as_grid:
                        if not row:  # Skip empty rows.
                            continue
                        key = row[0].strip()
                        if key and key in gid_to_data[gid]:
                            # The key is already in use.
                            raise base.PreprocessorError(
                                'Duplicate key in use in sheet {}: {}'.format(gid, key))
                        if key and not key.startswith(IGNORE_INITIAL):
                            # Grids use the first column as the key and make
                            # object out of the remaining columns.
                            grid_obj = {}
                            row = row[1:]
                            row_len = len(row)
                            for col, grid_key in enumerate(headers):
                                if not grid_key or grid_key.startswith(IGNORE_INITIAL):
                                    continue
                                value = (row[col] if row_len >
                                         col else '').strip()
                                if value or keep_empty_values:
                                    grid_obj[grid_key] = value
                            gid_to_data[gid][key] = grid_obj
                    elif format_as_map:
                        if not row:  # Skip empty rows.
                            continue
                        key = row[0].strip()
                        if not key and generate_ids:
                            key = 'untranslated_{}'.format(generated_key_index)
                            generated_key_index += 1
                        if key and not key.startswith(IGNORE_INITIAL):
                            if format_as == 'strings' and '@' not in key:
                                key = '{}@'.format(key)
                            value = row[1] if len(row) == 2 else ''
                            # Ignore empty values.
                            if value or keep_empty_values:
                                gid_to_data[gid][key] = value
                    else:
                        row_values = {}
                        for idx, column in enumerate(headers):
                            if not column.startswith(IGNORE_INITIAL):
                                value = row[idx] if len(row) > idx else ''
                                # Ignore empty values.
                                if value or keep_empty_values:
                                    row_values[column] = value
                        gid_to_data[gid].append(row_values)

        return gid_to_sheet, gid_to_data

    @staticmethod
    def parse_path(path):
        if ':' in path:
            return path.rsplit(':', 1)
        return path, None

    def _maybe_preserve_content(self, new_data, path, key_to_update, properties):
        # Includes meta properties from the Google Sheet.
        if self.config.include_properties:
            if META_KEY not in new_data:
                new_data[META_KEY] = {}
            if 'properties' not in new_data[META_KEY]:
                new_data[META_KEY]['properties'] = {}
            for name in self.config.include_properties:
                if name in properties:
                    new_data[META_KEY]['properties'][name] = properties[name]
        # Tabs colored red are marked draft.
        if isinstance(new_data, dict) \
                and self.config.color_as_draft \
                and properties.get('tabColor'):
            if properties['tabColor'] == {'red': 1}:
                new_data[DRAFT_KEY] = True
        if path.endswith(('.yaml', '.yml')) and self.config.preserve:
            # Use existing data if it exists. If we're updating data at a
            # specific key, and if the existing data doesn't exist, use an
            # empty dict. If the file doesn't exist and if we're not updating
            # at a specific key, just return the new data without reformatting.
            if self.pod.file_exists(path):
                # Do a text parse of the yaml file to prevent the constructors.
                content = self.pod.read_file(path)
                existing_data = utils.load_plain_yaml(content)
            elif key_to_update:
                existing_data = {}
            else:
                return new_data
            # Skip trying to update lists, because there would be no
            # expectation of merging old and new list data.
            if not key_to_update and not isinstance(new_data, dict):
                return new_data
            if isinstance(existing_data, dict):
                return utils.format_existing_data(
                    old_data=existing_data, new_data=new_data,
                    preserve=self.config.preserve, key_to_update=key_to_update)
        return new_data

    def execute(self, config):
        spreadsheet_id = config.id
        gids = config.gids or []
        if config.gid is not None:
            gids.append(config.gid)
        if not gids and not config.collection:
            gids.append(0)
        format_as = config.format
        keep_empty_values = config.keep_empty_values
        if (config.collection and
                format_as not in GoogleSheetsPreprocessor.MAP_TYPES and
                format_as not in GoogleSheetsPreprocessor.GRID_TYPES):
            format_as = 'map'
        gid_to_sheet, gid_to_data = GoogleSheetsPreprocessor.download(
            spreadsheet_id=spreadsheet_id, gids=gids, format_as=format_as,
            logger=self.pod.logger, generate_ids=config.generate_ids,
            header_row_count=config.header_row_count,
            header_row_index=config.header_row_index,
            keep_empty_values=keep_empty_values)

        if config.path:
            # Single sheet import.
            path, key_to_update = self.parse_path(config.path)

            for gid in gids:
                if gid not in gid_to_data:
                    self.logger.info(
                        'Sheet not imported for gid {}. Skipped tab?'.format(gid))
                    continue
                gid_to_data[gid] = self._maybe_preserve_content(
                    new_data=gid_to_data[gid],
                    path=path,
                    key_to_update=key_to_update,
                    properties=gid_to_sheet[gid])
                content = GoogleSheetsPreprocessor.serialize_content(
                    formatted_data=gid_to_data[gid], path=path,
                    output_style=self.config.output_style)

                self.pod.write_file(path, content)
                self.logger.info(
                    'Downloaded {} ({}) -> {}'.format(
                        gid_to_sheet[gid]['title'], gid, path))
        else:
            # Multi sheet import based on collection.
            collection_path = config.collection

            if not gids:
                gids = list(gid_to_sheet.keys())

            num_saved = 0
            for gid in gids:
                if gid not in gid_to_data:
                    self.logger.info(
                        'Sheet not imported for gid {}. Skipped tab?'.format(gid))
                    continue
                title = gid_to_sheet[gid]['title']
                if title.strip().startswith(IGNORE_INITIAL):
                    continue
                if self.pod.is_enabled(self.pod.FEATURE_OLD_SLUGIFY):
                    slug = utils.slugify(title)
                else:
                    slug = slugify.slugify(title)
                file_name = '{}.yaml'.format(slug)
                output_path = os.path.join(collection_path, file_name)
                gid_to_data[gid] = self._maybe_preserve_content(
                    new_data=gid_to_data[gid],
                    path=output_path,
                    key_to_update=None,
                    properties=gid_to_sheet[gid])
                # Use plain text dumper to preserve yaml constructors.
                output_content = utils.dump_plain_yaml(gid_to_data[gid])
                self.pod.write_file(output_path, output_content)
                if gid_to_data[gid].get(DRAFT_KEY):
                    self.logger.info('Drafted tab -> {}'.format(title))
                num_saved += 1
            text = 'Saved {} tabs -> {}'
            self.logger.info(text.format(num_saved, collection_path))

    @classmethod
    def get_convert_to(cls, path):
        ext = os.path.splitext(path)[1]
        convert_to = None
        if ext == '.json':
            return ext
        elif ext in ['.yaml', '.yml']:
            return ext
        return convert_to

    @classmethod
    def format_content(cls, content, path, format_as=None, preserve=None,
                       existing_data=None, key_to_update=None):
        """Formats content into either a CSV (text), list, or dictionary."""
        convert_to = cls.get_convert_to(path)
        if convert_to in ['.json', '.yaml', '.yml']:
            fp = io.StringIO()
            fp.write(content)
            fp.seek(0)
            if format_as == 'map':
                formatted_data = GoogleSheetsPreprocessor.format_as_map(fp)
            else:
                reader = csv.DictReader(fp)
                formatted_data = list(reader)
            formatted_data = utils.format_existing_data(
                old_data=existing_data, new_data=formatted_data,
                preserve=preserve, key_to_update=key_to_update)
            return formatted_data
        return content

    @classmethod
    def serialize_content(cls, formatted_data, path, output_style=None):
        """Serializes an object into a string as JSON, YAML, or a CSV
        (default)."""
        kwargs = {}
        convert_to = cls.get_convert_to(path)
        if convert_to == '.json':
            if output_style == 'pretty':
                kwargs['indent'] = 2
                kwargs['separators'] = (',', ': ')
                kwargs['sort_keys'] = True
            return json.dumps(formatted_data, **kwargs)
        elif convert_to in ('.yaml', '.yml'):
            # Use plain text dumper to preserve yaml constructors.
            return utils.dump_plain_yaml(formatted_data)
        return formatted_data

    def can_inject(self, doc=None, collection=None):
        if not self.injected:
            return False
        path, key_to_update = self.parse_path(self.config.path)
        if doc and doc.pod_path == path:
            return True
        return False

    def _normalize_formatted_content(self, fields):
        # A hack that sends fields through a roundtrip json serialization to
        # avoid encoding issues with injected output from Google Sheets.
        fp = io.StringIO()
        json.dump(fields, fp)
        fp.seek(0)
        return json.load(fp)

    def inject(self, doc):
        spreadsheet_id = self.config.id
        gids = self.config.gids or []
        if self.config.gid is not None:
            gids.append(self.config.gid)
        format_as = self.config.format
        if self.config.collection and format_as not in self.MAP_TYPES:
            format_as = 'map'
        _, gid_to_data = GoogleSheetsPreprocessor.download(
            spreadsheet_id=spreadsheet_id, gids=gids, format_as=format_as,
            logger=self.pod.logger, generate_ids=self.config.generate_ids)

        if self.config.path:
            if format_as in ['list']:
                self.pod.logger.info(
                    'Cannot inject list formatted spreadsheet -> {}'.format(self.config.path))
                return
            # Single sheet import.
            path, key_to_update = self.parse_path(self.config.path)

            for gid in gids:
                # Preserve existing yaml data.
                if (path.endswith(('.yaml', '.yml'))
                        and self.config.preserve and self.pod.file_exists(path)):
                    existing_data = self.pod.read_yaml(path)
                    gid_to_data[gid] = utils.format_existing_data(
                        old_data=existing_data, new_data=gid_to_data[gid],
                        preserve=self.config.preserve, key_to_update=key_to_update)

                gid_to_data[gid] = untag.Untag.untag(gid_to_data[gid])

    def get_edit_url(self, doc=None):
        """Returns the URL to edit in Google Sheets."""
        gid = self.config.gid or '0'
        return GoogleSheetsPreprocessor._edit_url_format.format(
            id=self.config.id, gid=gid)



"""Grow preprocessors."""

import json
from protorpc import protojson
from grow.preprocessors import blogger
from grow.preprocessors import google_drive
from grow.preprocessors import gulp
from grow.preprocessors import webpack
from grow.extensions import extension_importer

_preprocessor_kinds_to_classes = {}

_builtins = (
    blogger.BloggerPreprocessor,
    google_drive.GoogleDocsPreprocessor,
    google_drive.GoogleSheetsPreprocessor,
    gulp.GulpPreprocessor,
    webpack.WebpackPreprocessor,
)


def register_preprocessor(class_obj):
    _preprocessor_kinds_to_classes[class_obj.KIND] = class_obj


def config_from_json(preprocessor_class, content):
    config_class = preprocessor_class.Config
    return protojson.decode_message(config_class, content)


def make_preprocessor(kind, config, pod):
    autorun = config.pop('autorun', True)
    name = config.pop('name', None)
    tags = config.pop('tags', None)
    inject = config.pop('inject', False)
    class_obj = _preprocessor_kinds_to_classes.get(kind)
    if class_obj is None:
        raise ValueError('No legacy preprocessor for "{}".'.format(kind))
    if isinstance(config, dict):
        config = json.dumps(config)
        config = config_from_json(class_obj, config)
    return class_obj(pod, config, autorun=autorun, name=name, tags=tags, inject=inject)


def register_builtins():
    for builtin in _builtins:
        register_preprocessor(builtin)


def register_extensions(extension_paths, pod_root):
    for path in extension_paths:
        cls = extension_importer.ExtensionImporter.find_extension(
            path, pod_root)
        register_preprocessor(cls)


register_builtins()
